import logging
import os
import time
import inspect
import functools
import configparser
from os import PathLike
from typing import Union, Optional, Any
import attr
import click
import colorlog  # type: ignore
from click_skeleton.helpers import str2bool, seconds_to_human

logger = logging.getLogger(__name__)

VERBOSITIES = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
}

DEFAULT_CHECK_VERSION = False
DEFAULT_CONFIG = '~/musicbot.ini'
DEFAULT_LOG = ''
DEFAULT_DEBUG = False
DEFAULT_INFO = False
DEFAULT_WARNING = False
DEFAULT_ERROR = False
DEFAULT_CRITICAL = False
DEFAULT_TIMINGS = False
DEFAULT_VERBOSITY = 'warning'
DEFAULT_QUIET = False

MB_CHECK_VERSION = 'MB_CHECK_VERSION'
MB_CONFIG = 'MB_CONFIG'
MB_LOG = 'MB_LOG'
MB_DEBUG = 'MB_DEBUG'
MB_INFO = 'MB_INFO'
MB_WARNING = 'MB_WARNING'
MB_ERROR = 'MB_ERROR'
MB_CRITICAL = 'MB_CRITICAL'
MB_TIMINGS = 'MB_TIMINGS'
MB_VERBOSITY = 'MB_VERBOSITY'
MB_QUIET = 'MB_QUIET'


@attr.s(auto_attribs=True)
class Config:
    log: Optional[Union[str, PathLike]] = DEFAULT_LOG
    check_version: bool = DEFAULT_CHECK_VERSION
    quiet: bool = DEFAULT_QUIET
    debug: bool = DEFAULT_DEBUG
    info: bool = DEFAULT_INFO
    warning: bool = DEFAULT_WARNING
    error: bool = DEFAULT_ERROR
    critical: bool = DEFAULT_CRITICAL
    timings: bool = DEFAULT_TIMINGS
    verbosity: str = DEFAULT_VERBOSITY
    config: str = DEFAULT_CONFIG
    level: int = VERBOSITIES[DEFAULT_VERBOSITY]

    def __attrs_post_init__(self) -> None:
        self.check_version = str2bool(os.getenv(MB_CHECK_VERSION, 'true'))

    def set(
        self,
        config: Optional[str] = None,
        debug: Optional[bool] = None,
        info: Optional[bool] = None,
        warning: Optional[bool] = None,
        error: Optional[bool] = None,
        critical: Optional[bool] = None,
        timings: Optional[bool] = None,
        quiet: Optional[bool] = None,
        verbosity: Optional[str] = None,
        log: Optional[Union[str, PathLike]] = None,
    ) -> None:
        self.config = config if config is not None else os.environ.get(MB_CONFIG, DEFAULT_CONFIG)
        self.quiet = quiet if quiet is not None else str2bool(os.environ.get(MB_QUIET, str(DEFAULT_QUIET)))
        self.debug = debug if debug is not None else str2bool(os.environ.get(MB_DEBUG, str(DEFAULT_DEBUG)))
        self.info = info if info is not None else str2bool(os.environ.get(MB_INFO, str(DEFAULT_INFO)))
        self.warning = warning if warning is not None else str2bool(os.environ.get(MB_WARNING, str(DEFAULT_WARNING)))
        self.error = error if error is not None else str2bool(os.environ.get(MB_ERROR, str(DEFAULT_ERROR)))
        self.critical = critical if critical is not None else str2bool(os.environ.get(MB_CRITICAL, str(DEFAULT_CRITICAL)))
        self.timings = timings if timings is not None else str2bool(os.environ.get(MB_TIMINGS, str(DEFAULT_TIMINGS)))
        self.verbosity = verbosity if verbosity is not None else os.environ.get(MB_VERBOSITY, DEFAULT_VERBOSITY)
        self.log = (log if log is not None else os.environ.get(MB_LOG, DEFAULT_LOG)) or None

        if self.debug:
            self.verbosity = 'debug'
        if self.info:
            self.verbosity = 'info'
        if self.warning:
            self.verbosity = 'warning'
        if self.error:
            self.verbosity = 'error'
        if self.critical:
            self.verbosity = 'critical'

        self.level = VERBOSITIES.get(self.verbosity, logging.WARNING)
        root_logger = logging.getLogger()
        root_logger.setLevel(self.level)
        handler = logging.StreamHandler()
        handler.setLevel(self.level)
        handler.setFormatter(
            colorlog.ColoredFormatter(
                fmt='%(log_color)s%(name)s | %(asctime)s | %(levelname)s | %(message)s',
                datefmt='%Y-%d-%d %H:%M:%S',
                log_colors={
                    'DEBUG': 'cyan',
                    'INFO': 'green',
                    'WARNING': 'yellow',
                    'ERROR': 'red',
                    'CRITICAL': 'red,bg_white',
                },
            )
        )
        root_logger.addHandler(handler)

        if self.log:
            fh = logging.FileHandler(self.log)
            fh.setLevel(logging.DEBUG)
            logging.getLogger().addHandler(fh)
            logger.debug(self)

    @functools.cached_property
    def configfile(self) -> configparser.ConfigParser:
        file = configparser.ConfigParser()
        file.read(self.config)
        valid_sections = ('musicbot', 'spotify')
        for section in valid_sections:
            if section not in file:
                logger.warning(f'[{section}] section is not present in {self.config}')
        return file

    def write(self) -> None:
        with open(self.config, 'w') as output_config:
            self.configfile.write(output_config)

    def timeit(self, *wrapper_args: Any, **wrapper_kwargs: Any) -> Any:
        func = None
        if len(wrapper_args) == 1 and callable(wrapper_args[0]):
            func = wrapper_args[0]
        if func:
            always = False
            on_success = False
            on_config = True
        else:
            always = wrapper_kwargs.get('always', False)
            on_success = wrapper_kwargs.get('on_success', False)
            on_config = wrapper_kwargs.get('on_config', True)

        def real_timeit(func: Any) -> Any:
            @functools.wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                start = int(time.time())
                result = None
                try:
                    result = func(*args, **kwargs)
                    return result
                finally:
                    end = int(time.time())
                    for_human = seconds_to_human(end - start)

                    if self.info or self.debug:
                        args_values = []
                        argspec = inspect.getfullargspec(func)
                        # go through each position based argument
                        counter = 0
                        if argspec.args and isinstance(argspec.args, list):
                            for arg in args:
                                # when you run past the formal positional arguments
                                try:
                                    args_values.append(str(argspec.args[counter]) + " = " + str(arg))
                                    counter += 1
                                except IndexError:
                                    # then fallback to using the positional varargs name
                                    if argspec.varargs:
                                        varargsname = argspec.varargs
                                        args_values.append("*" + varargsname + " = " + str(arg))

                        # finally show the named varargs
                        for k, v in kwargs.items():
                            args_values.append(k + " = " + str(v))

                        args_values_str = '|'.join(args_values)
                        if argspec.annotations.get('return', None) is None:
                            timing = f'(timings) {func.__module__}.{func.__qualname__} ({args_values_str}): {for_human}'
                        else:
                            timing = f'(timings) {func.__module__}.{func.__qualname__} ({args_values_str}): {for_human} | result = {result}'
                    else:
                        timing = f'(timings) {func.__module__}.{func.__qualname__} : {for_human}'
                    if always or (on_success and result) or (on_config and self.timings):
                        click.secho(timing, fg="magenta", bold=True, err=True)
            return wrapper
        return real_timeit(func) if func else real_timeit


config = Config()



from typing import Any
import requests
from click_option_group import optgroup  # type: ignore
from musicbot.config import config
from musicbot.graphql import GraphQL
from musicbot.helpers import config_string


DEFAULT_GRAPHQL_ADMIN = 'http://127.0.0.1:5001/graphql'
graphql_admin_option = [
    optgroup.option(
        '--graphql-admin',
        help='GraphQL endpoint',
        default=DEFAULT_GRAPHQL_ADMIN,
        callback=config_string,
        show_default=True
    )
]


DEFAULT_GRAPHQL_ADMIN_USER = None
graphql_admin_user_option = [
    optgroup.option(
        '--graphql-admin-user',
        help='GraphQL admin user (basic auth)',
        default=DEFAULT_GRAPHQL_ADMIN_USER,
        callback=config_string,
    )
]


DEFAULT_GRAPHQL_ADMIN_PASSWORD = None
graphql_admin_password_option = [
    optgroup.option(
        '--graphql-admin-password',
        help='GraphQL admin password (basic auth)',
        default=DEFAULT_GRAPHQL_ADMIN_PASSWORD,
        callback=config_string,
    )
]

admin_options =\
    [optgroup.group('Admin options')] +\
    graphql_admin_option +\
    graphql_admin_user_option +\
    graphql_admin_password_option


class Admin(GraphQL):
    @config.timeit
    def __init__(self, graphql: str, graphql_admin_user: str, graphql_admin_password: str) -> None:
        self.user = graphql_admin_user
        self.password = graphql_admin_password
        GraphQL.__init__(
            self,
            graphql=graphql,
            authorization=requests.auth._basic_auth_str(self.user, self.password)
        )

    @config.timeit
    def users(self) -> Any:
        query = """{
            accountsList {
                user {
                  lastName
                  createdAt
                  updatedAt
                  firstName
                }
                email
            }
        }"""
        return self.post(query)['data']['accountsList']



import logging
import operator
import textwrap
import json
import shutil
import collections
import click
import jellyfish  # type: ignore
from prettytable import PrettyTable  # type: ignore
from slugify import slugify  # type: ignore
from colorama import Fore  # type: ignore
from click_skeleton import AdvancedGroup, add_options

from musicbot import helpers
from musicbot.cli import main_cli
from musicbot.spotify import spotify_options
from musicbot.user import auth_options

logger = logging.getLogger(__name__)


def dump_tracks(tracks):
    tracks = [
        {
            'title': t['track']['name'],
            'artist': t['track']['artists'][0]['name'],
            'album': t['track']['album']['name'],
        } for t in tracks
    ]
    print(json.dumps(tracks))


def print_tracks_table(tracks):
    if not tracks:
        return
    pt = PrettyTable(["Track", "Artist", "Album"])
    width = shutil.get_terminal_size().columns // 3
    for t in tracks:
        title = '\n'.join(textwrap.wrap(t['track']['name'], width))
        artist = '\n'.join(textwrap.wrap(t['track']['artists'][0]['name'], width))
        album = '\n'.join(textwrap.wrap(t['track']['album']['name'], width))
        pt.add_row([title, artist, album])
    print(pt)


def print_distances(distances):
    if not distances:
        return
    pt = PrettyTable(["Title", "Artist", "Album", "Distance"])
    for distance in distances:
        st = distance['spotify_track']
        stitle = st['track']['name']
        sartist = st['track']['artists'][0]['name']
        salbum = st['track']['album']['name']
        lt = distance['local_track']
        d = distance['distance']
        pt.add_row([
            f"{Fore.GREEN}Spotify : {stitle}{Fore.RESET}\nLocal : {lt['title']}",
            f"{Fore.GREEN}Spotify : {sartist}{Fore.RESET}\n Local : {lt['artist']}",
            f"{Fore.GREEN}Spotify : {salbum}{Fore.RESET}\nLocal : {lt['album']}",
            d,
        ])
    print(pt)


def print_playlists_table(playlists):
    if not playlists:
        return
    pt = PrettyTable(["Name", "Size"])
    for p in playlists:
        pt.add_row([p['name'], p['tracks']['total']])
    print(pt.get_string(title='Spotify playlists'))


def output_tracks(output: str, tracks):
    if output == 'table':
        print_tracks_table(tracks)
    elif output == 'json':
        dump_tracks(tracks)


@main_cli.group('spotify', help='Spotify tool', cls=AdvancedGroup)
def spotify_cli():
    pass


@spotify_cli.command(help='List playlists')
@add_options(spotify_options)
def playlists(spotify):
    print_playlists_table(spotify.playlists())


@spotify_cli.command(help='Show playlist')
@add_options(
    spotify_options +
    helpers.output_option
)
@click.argument("name")
def playlist(name, spotify, output):
    tracks = spotify.playlist(name)
    output_tracks(output, tracks)


@spotify_cli.command(help='Show tracks')
@add_options(
    spotify_options +
    helpers.output_option
)
def tracks(spotify, output):
    tracks = spotify.tracks()
    output_tracks(output, tracks)


@spotify_cli.command(help='Diff between local and spotify')
@add_options(
    auth_options +
    spotify_options +
    helpers.output_option
)
@click.option('--min-threshold', help='Minimum distance threshold', type=click.FloatRange(0.0, 1.0), default=0.9)
@click.option('--max-threshold', help='Maximum distance threshold', type=click.FloatRange(0.0, 1.0), default=1.0)
def diff(user, spotify, output, min_threshold, max_threshold):
    stopwords = ['the', 'remaster', 'remastered', 'cut', 'part', 'version', 'mix', 'deluxe', 'edit', 'album', 'lp'] + list(map(str, range(1900, 2020)))
    replacements = [['praxis', 'buckethead'], ['lawson-rollins', 'buckethead']]

    spotify_tracks = spotify.tracks()
    local_tracks = user.do_filter()

    spotify_tracks = {
        slugify(f"""{t['track']['artists'][0]['name']}-{t['track']['name']}""", stopwords=stopwords, replacements=replacements):
        t for t in spotify_tracks
    }

    local_tracks = {
        slugify(f"""{t['artist']}-{t['title']}""", stopwords=stopwords, replacements=replacements):
        t for t in local_tracks
    }

    spotify_differences = set(spotify_tracks.keys()).difference(set(local_tracks.keys()))
    spotify_slug_tracks = collections.OrderedDict((d, spotify_tracks[d]) for d in sorted(spotify_differences))

    output_tracks(output, spotify_slug_tracks.values())
    print(f"found in local     : {len(local_tracks) - len(spotify_differences)}")
    print(f"not found in local : {len(spotify_differences)}")

    distances_tracks = []
    for spotify_slug, spotify_track in spotify_slug_tracks.items():
        distances = {
            local_slug: jellyfish.jaro_similarity(spotify_slug, local_slug)
            for local_slug in local_tracks
        }
        closest_local_track = max(distances.items(), key=operator.itemgetter(1))
        closest_local_slug = closest_local_track[0]
        closest_distance = closest_local_track[1]

        if min_threshold <= closest_distance <= max_threshold:
            if 'spotify-error' in local_tracks[closest_local_slug]['keywords']:
                continue
            distances_tracks.append({
                'local_track': local_tracks[closest_local_slug],
                'local_slug': closest_local_slug,
                'spotify_track': spotify_track,
                'spotify_slug': spotify_slug,
                'distance': closest_distance,
            })
    print_distances(distances_tracks)



import logging
import json
from typing import Any, Dict, List, Optional
import attr
import click
from click_option_group import optgroup  # type: ignore

logger = logging.getLogger(__name__)


rating_choices = [x * 0.5 for x in range(0, 11)]


def sane_rating(ctx: click.Context, param: click.ParamType, value: float) -> float:
    if value is not None and value in rating_choices:
        ctx.params[param.name] = value
        return value
    return float('nan')


min_int = 0
max_int = 2147483647

default_name: Optional[str] = None
default_filter = None
default_relative = False
default_shuffle = False
default_youtubes: List[str] = []
default_no_youtubes: List[str] = []
default_spotifys: List[str] = []
default_no_spotifys: List[str] = []
default_formats: List[str] = []
default_no_formats: List[str] = []
default_genres: List[str] = []
default_no_genres: List[str] = []
default_limit: int = max_int
default_min_duration: int = min_int
default_max_duration: int = max_int
default_min_size: int = min_int
default_max_size: int = max_int
default_min_rating = min(rating_choices)
default_max_rating = max(rating_choices)
default_keywords: List[str] = []
default_no_keywords: List[str] = []
default_artists: List[str] = []
default_no_artists: List[str] = []
default_titles: List[str] = []
default_no_titles: List[str] = []
default_albums: List[str] = []
default_no_albums: List[str] = []


@attr.s(auto_attribs=True, frozen=True)
class Filter:
    name: Optional[str] = default_name
    relative: bool = default_relative
    shuffle: bool = default_shuffle
    min_duration: int = default_min_duration
    max_duration: int = default_max_duration
    min_size: int = default_min_size
    max_size: int = default_max_size
    min_rating: float = default_min_rating
    max_rating: float = default_max_rating
    youtubes: List[str] = default_youtubes
    no_youtubes: List[str] = default_no_youtubes
    spotifys: List[str] = default_spotifys
    no_spotifys: List[str] = default_no_spotifys
    formats: List[str] = default_formats
    no_formats: List[str] = default_no_formats
    limit: int = default_limit
    genres: List[str] = default_genres
    no_genres: List[str] = default_no_genres
    keywords: List[str] = default_keywords
    no_keywords: List[str] = default_no_keywords
    artists: List[str] = default_artists
    no_artists: List[str] = default_no_artists
    titles: List[str] = default_titles
    no_titles: List[str] = default_no_titles
    albums: List[str] = default_albums
    no_albums: List[str] = default_no_albums

    def __attrs_post_init__(self) -> None:
        if self.min_size > self.max_size:
            raise ValueError(f"Invalid minimum ({self.min_rating}) or maximum ({self.max_rating}) size")

        if self.min_rating not in rating_choices:
            raise ValueError(f"Invalid minimum rating {self.min_rating}, it should be one of {rating_choices}")

        if self.max_rating not in rating_choices:
            raise ValueError(f"Invalid maximum rating {self.max_rating}, it should be one of {rating_choices}")

        if self.min_rating > self.max_rating:
            raise ValueError(f"Invalid minimum ({self.min_rating}) or maximum ({self.max_rating}) rating")

        if self.min_duration > self.max_duration:
            raise ValueError(f"Invalid minimum ({self.min_duration}) or maximum ({self.max_duration}) duration")

        is_bad_formats = set(self.formats).intersection(self.no_formats)
        is_bad_artists = set(self.artists).intersection(self.no_artists)
        is_bad_genres = set(self.genres).intersection(self.no_genres)
        is_bad_albums = set(self.albums).intersection(self.no_albums)
        is_bad_titles = set(self.titles).intersection(self.no_titles)
        is_bad_keywords = set(self.keywords).intersection(self.no_keywords)
        not_empty_set = is_bad_formats or is_bad_artists or is_bad_genres or is_bad_albums or is_bad_titles or is_bad_keywords
        if not_empty_set:
            raise ValueError(f"You can't have duplicates value in filters {self}")
        logger.debug(f'Filter: {self}')

    def __repr__(self) -> str:
        return json.dumps(self.as_dict())

    def diff(self) -> Dict[str, Any]:
        '''Print only differences with default filter'''
        myself = vars(self)
        default = vars(Filter())
        return {k: myself[k] for k in myself if default[k] != myself[k] and k != 'name'}

    def common(self) -> Dict[str, Any]:
        '''Print common values with default filter'''
        myself = vars(self)
        default = vars(Filter())
        return {k: myself[k] for k in myself if default[k] == myself[k] and k != 'name'}

    def as_dict(self) -> Dict[str, Any]:  # pylint: disable=unsubscriptable-object
        return {
            'minDuration': self.min_duration,
            'maxDuration': self.max_duration,
            'minSize': self.min_size,
            'maxSize': self.max_size,
            'minRating': self.min_rating,
            'maxRating': self.max_rating,
            'artists': self.artists,
            'noArtists': self.no_artists,
            'albums': self.albums,
            'noAlbums': self.no_albums,
            'titles': self.titles,
            'noTitles': self.no_titles,
            'genres': self.genres,
            'noGenres': self.no_genres,
            'formats': self.formats,
            'noFormats': self.no_formats,
            'keywords': self.keywords,
            'noKeywords': self.no_keywords,
            'shuffle': self.shuffle,
            'relative': self.relative,
            'limit': self.limit,
            'youtubes': self.youtubes,
            'noYoutubes': self.no_youtubes,
            'spotifys': self.spotifys,
            'noSpotifys': self.no_spotifys,
        }

    def to_graphql(self) -> str:
        return ", ".join([f'{k}: {json.dumps(v)}' for k, v in self.as_dict().items()])


options = [
    optgroup.option(
        '--name',
        help='Filter name',
        default=default_name,
        is_eager=True,
    ),
    optgroup.option(
        '--limit',
        help='Fetch a maximum limit of music',
        default=default_limit,
        is_eager=True,
    ),
    optgroup.option(
        '--youtubes',
        help='Select musics with a youtube link',
        multiple=True,
        default=default_youtubes,
        is_eager=True,
    ),
    optgroup.option(
        '--no-youtubes',
        help='Select musics without youtube link',
        multiple=True,
        default=default_no_youtubes,
        is_eager=True,
    ),
    optgroup.option(
        '--spotifys',
        help='Select musics with a spotifys link',
        multiple=True,
        default=default_youtubes,
        is_eager=True,
    ),
    optgroup.option(
        '--no-spotifys',
        help='Select musics without spotifys link',
        multiple=True,
        default=default_no_youtubes,
        is_eager=True,
    ),
    optgroup.option(
        '--formats',
        help='Select musics with file format',
        multiple=True,
        default=default_formats,
        is_eager=True,
    ),
    optgroup.option(
        '--no-formats',
        help='Filter musics without format',
        multiple=True,
        default=default_no_formats,
        is_eager=True,
    ),
    optgroup.option(
        '--keywords',
        help='Select musics with keywords',
        multiple=True,
        default=default_keywords,
        is_eager=True,
    ),
    optgroup.option(
        '--no-keywords',
        help='Filter musics without keywords',
        multiple=True,
        default=default_no_keywords,
        is_eager=True,
    ),
    optgroup.option(
        '--artists',
        help='Select musics with artists',
        multiple=True,
        default=default_artists,
        is_eager=True,
    ),
    optgroup.option(
        '--no-artists',
        help='Filter musics without artists',
        multiple=True,
        default=default_no_artists,
        is_eager=True,
    ),
    optgroup.option(
        '--albums',
        help='Select musics with albums',
        multiple=True,
        default=default_albums,
        is_eager=True,
    ),
    optgroup.option(
        '--no-albums',
        help='Filter musics without albums',
        multiple=True,
        default=default_no_albums,
        is_eager=True,
    ),
    optgroup.option(
        '--titles',
        help='Select musics with titles',
        multiple=True,
        default=default_titles,
        is_eager=True,
    ),
    optgroup.option(
        '--no-titles',
        help='Filter musics without titless',
        multiple=True,
        default=default_no_titles,
        is_eager=True,
    ),
    optgroup.option(
        '--genres',
        help='Select musics with genres',
        multiple=True,
        default=default_genres,
        is_eager=True,
    ),
    optgroup.option(
        '--no-genres',
        help='Filter musics without genres',
        multiple=True,
        default=default_no_genres,
        is_eager=True,
    ),
    optgroup.option(
        '--min-duration',
        help='Minimum duration filter (hours:minutes:seconds)',
        default=default_min_duration,
        is_eager=True,
    ),
    optgroup.option(
        '--max-duration',
        help='Maximum duration filter (hours:minutes:seconds))',
        default=default_max_duration,
        is_eager=True,
    ),
    optgroup.option(
        '--min-size',
        help='Minimum file size filter (in bytes)',
        default=default_min_size,
        is_eager=True,
    ),
    optgroup.option(
        '--max-size',
        help='Maximum file size filter (in bytes)',
        default=default_max_size,
        is_eager=True,
    ),
    optgroup.option(
        '--min-rating',
        help='Minimum rating',
        default=default_min_rating,
        show_default=True,
        callback=sane_rating,
        is_eager=True,
    ),
    optgroup.option(
        '--max-rating',
        help='Maximum rating',
        default=default_max_rating,
        show_default=True,
        callback=sane_rating,
        is_eager=True,
    ),
    optgroup.option(
        '--relative',
        help='Generate relatives paths',
        default=default_relative,
        is_flag=True,
        is_eager=True,
    ),
    optgroup.option(
        '--shuffle',
        help='Randomize selection',
        default=default_shuffle,
        is_flag=True,
        is_eager=True,
    ),
]


def sane_filter(ctx: click.Context, param: click.ParamType, value: Any) -> Filter:  # pylint: disable=unused-argument
    kwargs = {}
    for field in attr.fields_dict(Filter):
        kwargs[field] = ctx.params[field]
        ctx.params.pop(field)

    myfilter = Filter(**kwargs)
    ctx.params['music_filter'] = myfilter
    return myfilter


DEFAULT_FILTER = None
mfilter_option = [
    optgroup.option(
        '--music-filter',
        help='Music Filter',
        expose_value=False,
        callback=sane_filter,
        hidden=True,
    )
]
mfilter_options =\
    [optgroup.group('Filter options')] +\
    mfilter_option +\
    options



import logging
import pathlib
import json
import os
from typing import Any, Optional, List, Dict, Iterable, Iterator
import click
import acoustid  # type: ignore
import mutagen  # type: ignore
from pydub import AudioSegment  # type: ignore
from click_option_group import optgroup  # type: ignore
from click_skeleton.helpers import mysplit
from .helpers import ensure


logger = logging.getLogger(__name__)

music_options_group = [optgroup.group('Music options')]
keywords_option = [optgroup.option('--keywords', help='Keywords', multiple=True)]
artist_option = [optgroup.option('--artist', help='Artist', default=None)]
album_option = [optgroup.option('--album', help='Album', default=None)]
title_option = [optgroup.option('--title', help='Title', default=None)]
genre_option = [optgroup.option('--genre', help='Genre', default=None)]
number_option = [optgroup.option('--number', help='Track number', default=None)]
rating_option = [optgroup.option('--rating', help='Rating', default=None)]

keywords_argument = [
    click.argument(
        'keywords',
        nargs=-1,
    ),
]

options =\
    music_options_group +\
    keywords_option +\
    artist_option +\
    album_option +\
    title_option +\
    genre_option +\
    number_option +\
    rating_option

DEFAULT_CHECKS = [
    'no-title',
    'no-artist',
    'no-album',
    'no-genre',
    'no-rating',
    'no-tracknumber',
    'invalid-title',
    'invalid-comment',
    'invalid-path',
]
path_argument = [
    click.argument(
        'path',
        type=click.Path(exists=True, dir_okay=False),
    ),
]
folder_option = [
    click.option(
        '--folder',
        help="Destination folder",
        type=click.Path(exists=True, file_okay=False),
    ),
]
folder_argument = [
    click.argument(
        'folder',
        type=click.Path(exists=True, file_okay=False),
    ),
]
checks_options = [
    optgroup.group('Check options'),
    optgroup.option(
        '--checks',
        help='Consistency tests',
        multiple=True,
        default=DEFAULT_CHECKS,
        show_default=True,
        type=click.Choice(DEFAULT_CHECKS),
    ),
    optgroup.option(
        '--fix',
        help="Fix musics",
        is_flag=True,
    ),
]
supported_formats = ["mp3", "flac"]


# pylint: disable-msg=unsupported-membership-test
# pylint: disable-msg=unsubscriptable-object
# pylint: disable-msg=unsupported-assignment-operation
class File:
    def __init__(self, path: str, folder: Optional[str] = None, youtube_link: Optional[str] = None, spotify_link: Optional[str] = None):
        self.path = path
        self.folder = folder if folder is not None else ''
        self.youtube_link = youtube_link if youtube_link is not None else ''
        self.spotify_link = spotify_link if spotify_link is not None else ''
        self.handle = mutagen.File(self.path)
        self.inconsistencies = []
        if not self.title:
            self.inconsistencies.append("no-title")
        if self.genre == '':
            self.inconsistencies.append("no-genre")
        if self.album == '':
            self.inconsistencies.append("no-album")
        if self.artist == '':
            self.inconsistencies.append("no-artist")
        if self.rating == -1:
            self.inconsistencies.append("no-rating")
        if self.number == -1:
            self.inconsistencies.append("no-tracknumber")
        if self.extension == '.flac' and self._comment and not self._description:
            self.inconsistencies.append('invalid-comment')
        if self.extension == '.mp3' and self._description and not self._comment:
            self.inconsistencies.append('invalid-comment')
        if self.number not in (-1, 0) and self.title != self.canonic_title:
            logger.info(f"invalid-title, '{self.title}' should be '{self.canonic_title}'")
            self.inconsistencies.append("invalid-title")
        if self.number not in (-1, 0) and not self.path.endswith(self.canonic_path):
            logger.info(f"invalid-path, '{self.path}' should be '{self.canonic_path}'")
            self.inconsistencies.append("invalid-path")

    def __repr__(self) -> str:
        return self.path

    def __iter__(self) -> Iterator[Any]:
        yield from self.as_dict().items()

    def close(self) -> None:
        self.handle.close()

    def to_mp3(self, folder: Optional[str] = None, dry: Optional[bool] = None) -> bool:
        dry = dry if dry is not None else False
        folder = folder if folder is not None else self.folder

        if self.extension != '.flac':
            logger.error(f"{self} is not a flac file")
            return False
        if 'invalid-path' in self.inconsistencies:
            logger.error(f"{self} does not have a canonic path like : {self.canonic_artist_album_filename}")
            return False

        mp3_path = os.path.join(folder, self.artist, self.album, self.canonic_title + '.mp3')
        if os.path.exists(mp3_path):
            logger.info(f"{mp3_path} already exists, not overwriting")
            return False
        logger.debug(f"{self} convert destination : {mp3_path}")
        if not dry:
            ensure(mp3_path)
            flac_audio = AudioSegment.from_file(self.path, "flac")
            flac_audio.export(mp3_path, format="mp3")
        return True

    def as_dict(self) -> Dict[str, Any]:  # pylint: disable=unsubscriptable-object
        return {
            'title': self.title,
            'album': self.album,
            'genre': self.genre,
            'artist': self.artist,
            'folder': self.folder,
            'youtube': self.youtube,
            'spotify': self.spotify,
            'number': self.number,
            'path': self.path,
            'rating': self.rating,
            'duration': self.duration,
            'size': self.size,
            'keywords': self.keywords,
        }

    @property
    def extension(self) -> str:
        return pathlib.Path(self.path).suffix

    def to_graphql(self) -> str:
        return ", ".join([f'{k}: {json.dumps(v)}' for k, v in self.as_dict().items()])

    def to_json(self) -> str:
        return json.dumps(self.as_dict())

    @property
    def canonic_path(self) -> str:
        return str(pathlib.PurePath(self.folder, self.canonic_artist_album_filename))

    @property
    def canonic_artist_album_filename(self) -> str:
        return str(pathlib.PurePath(self.artist, self.album, self.canonic_filename))

    @property
    def filename(self) -> str:
        return os.path.basename(self.path)

    @property
    def canonic_filename(self) -> str:
        return f"{self.canonic_title}{self.extension}"

    def _get_first(self, tag: str, default: str = '') -> str:
        if tag not in self.handle:
            return default
        for item in self.handle[tag]:
            return str(item)
        return default

    @property
    def duration(self) -> int:
        return int(self.handle.info.length)

    @property
    def size(self) -> int:
        return os.path.getsize(self.path)

    @property
    def title(self) -> str:
        if self.extension == '.flac':
            return self._get_first('title')
        return self._get_first('TIT2')

    @title.setter
    def title(self, title: str) -> None:
        if self.extension == '.flac':
            self.handle.tags['title'] = title
        else:
            self.handle.tags.add(mutagen.id3.TIT2(text=title))

    @property
    def canonic_title(self) -> str:
        return f"{str(self.number).zfill(2)} - {self.title}"

    @property
    def album(self) -> str:
        if self.extension == '.flac':
            return self._get_first('album')
        return self._get_first('TALB')

    @album.setter
    def album(self, album: str) -> None:
        if self.extension == '.flac':
            self.handle.tags['album'] = album
        else:
            self.handle.tags.add(mutagen.id3.TALB(text=album))

    @property
    def artist(self) -> str:
        if self.extension == '.flac':
            return self._get_first('artist')
        return self._get_first('TPE1')

    @artist.setter
    def artist(self, artist: str) -> None:
        if self.extension == '.flac':
            self.handle.tags['artist'] = artist
        else:
            self.handle.tags.add(mutagen.id3.TPE1(text=artist))

    @property
    def genre(self) -> str:
        if self.extension == '.flac':
            return self._get_first('genre')
        return self._get_first('TCON')

    @genre.setter
    def genre(self, genre: str) -> None:
        if self.extension == '.flac':
            self.handle.tags['genre'] = genre
        else:
            self.handle.tags.add(mutagen.id3.TCON(text=genre))

    @property
    def rating(self) -> float:
        if self.extension == '.flac':
            s = self._get_first('fmps_rating')
        else:
            s = self._get_first('TXXX:FMPS_Rating')
        try:
            n = float(s)
            if n < 0.0:
                return 0.0
            return n * 5.0
        except ValueError:
            return -1

    @rating.setter
    def rating(self, rating: float) -> None:
        if self.extension == '.flac':
            self.handle['fmps_rating'] = str(rating)
        else:
            self.handle.tags.add(mutagen.id3.TXXX(desc='FMPS_Rating', text=str(rating)))

    @property
    def _comment(self) -> str:
        return self._get_first('COMM:ID3v1 Comment:eng')

    @_comment.setter
    def _comment(self, comment: str) -> None:
        self.handle.tags.add(mutagen.id3.COMM(desc='ID3v1 Comment:eng', text=str(comment)))

    @property
    def _description(self) -> str:
        return self._get_first('description')

    @_description.setter
    def _description(self, description: str) -> None:
        self.handle.tags['description'] = description

    @property
    def number(self) -> int:
        if self.extension == '.flac':
            s = self._get_first('tracknumber')
        else:
            s = self._get_first('TRCK')
        try:
            if '/' in s:
                s = s.split('/')[0]
            n = int(s)
            if n < 0:
                return -1
            if n > 2 ** 31 - 1:
                logger.warning(f"{self} : invalid track number {n}")
                return 0
            return n
        except ValueError:
            return -1

    @number.setter
    def number(self, number: int) -> None:
        if self.extension == '.flac':
            self.handle.tags['tracknumber'] = str(number)
        else:
            self.handle.tags.add(mutagen.id3.TRCK(text=str(number)))

    @property
    def keywords(self) -> List[str]:
        if self.extension == '.mp3':
            return mysplit(self._comment, ' ')
        if self.extension == '.flac':
            if self._comment and not self._description:
                logger.warning(f'{self} : fixing flac keywords with mp3 comment')
                self._description = self._comment
            return mysplit(self._description, ' ')
        return []

    @keywords.setter
    def keywords(self, keywords: Iterable[str]) -> None:
        if self.extension == '.mp3':
            logger.info(f'{self} : mp3 {keywords}')
            self._comment = ' '.join(keywords)
        elif self.extension == '.flac':
            logger.info(f'{self} : flac {keywords}')
            self._description = ' '.join(keywords)
        else:
            logger.error(f'{self} : unknown extension {self.extension}')

    def add_keywords(self, keywords: Iterable[str], dry: Optional[bool] = None) -> bool:
        dry = dry if dry is not None else False
        tags = self.keywords
        for k in keywords:
            if k not in tags:
                logger.info(f'{self} : new keyword {k}')
                tags.append(k)
        if set(self.keywords) != set(tags):
            self.keywords = tags
            print(self.keywords)
            print(tags)
            logger.info(f'{self} : new keywords {self.keywords}')
            return self.save(dry)
        return False

    def delete_keywords(self, keywords: Iterable[str], dry: Optional[bool] = None) -> bool:
        dry = dry if dry is not None else False
        tags = self.keywords
        for k in keywords:
            if k in tags:
                logger.info(f'{self} : new keyword {k}')
                tags.remove(k)
        if set(self.keywords) != set(tags):
            self.keywords = tags
            logger.info(f'{self} : new keywords {self.keywords}')
            return self.save(dry)
        return False

    @property
    def youtube(self) -> str:
        return self.youtube_link

    @property
    def spotify(self) -> str:
        return self.spotify_link

    def fingerprint(self, api_key: str) -> str:
        ids = acoustid.match(api_key, self.path)
        for score, recording_id, title, artist in ids:
            logger.info(f"{self} score : {score} | recording_id : {recording_id} | title : {title} | artist : {artist}")
            return str(recording_id)
        logger.info(f'{self} : fingerprint cannot be detected')
        return ''

    def fix(self, checks: Optional[Iterable[str]] = None, dry: Optional[bool] = None) -> bool:
        checks = checks if checks is not None else DEFAULT_CHECKS
        dry = dry if dry is not None else False
        if 'no-rating' in checks:
            self.rating = 0.0
            self.inconsistencies.remove('no-rating')
        if 'no-tracknumber' in checks:
            self.number = 0
            self.inconsistencies.remove('no-tracknumber')
        if 'invalid-comment' in checks:
            if self.extension == '.flac' and self._comment and not self._description:
                self._description = ''
            if self.extension == '.mp3' and self._description and not self._comment:
                self._comment = ''
            self.inconsistencies.remove('invalid-comment')
        if 'invalid-title' in checks:
            logger.info(f"{self} : {self.title} => {self.canonic_title}")
            self.title = self.canonic_title
            self.inconsistencies.remove('invalid-title')
        if 'invalid-path' in checks:
            logger.info(f"{self} : {self.path} => {self.canonic_path}")
            if not dry:
                ensure(self.canonic_path)
                path = pathlib.Path(self.path)
                path.replace(self.canonic_path)
                self.path = self.canonic_path
                self.handle = mutagen.File(self.path)
                self.inconsistencies.remove('invalid-path')
        return self.save(dry)

    def save(self, dry: Optional[bool] = None) -> bool:
        dry = dry if dry is not None else False
        try:
            if not dry:
                self.handle.save()
            return True
        except mutagen.MutagenError as e:
            logger.error(e)
        return False



import click
from musicbot.helpers import config_string

DEFAULT_ACOUSTID_API_KEY = None
acoustid_api_key_option = [
    click.option(
        '--acoustid-api-key',
        help='AcoustID API Key',
        default=DEFAULT_ACOUSTID_API_KEY,
        callback=config_string,
    )
]



import logging
import click
from logging_tree import printout  # type: ignore
from click_skeleton import AdvancedGroup
from musicbot.cli import main_cli
from musicbot.config import config as config_obj

logger = logging.getLogger('musicbot')


@main_cli.group('config', hidden=True, help='Config management', cls=AdvancedGroup)
def config_cli():
    pass


@config_cli.command(help='Print global config')
def show():
    print(config_obj)


@config_cli.command(help='Write in global config')
@click.argument('section')
@click.argument('key')
@click.argument('value')
def write(section, key, value):
    if not config_obj.configfile.has_section(section):
        config_obj.configfile.add_section(section)
    config_obj.configfile[section][key] = value
    config_obj.write()


@config_cli.command('print', help='Print config file')
def _print():
    for each_section in config_obj.configfile.sections():
        print(f"[{each_section}]")
        for (each_key, each_val) in config_obj.configfile.items(each_section):
            print(f"    {each_key} = {each_val}")
        print()


@config_cli.command('logging', help='Print logging config')
def _logging():
    printout()



import sys
import os
import logging
from typing import Any, Collection, Iterable
import enlighten  # type: ignore
import click
from click_skeleton.helpers import mysplit
from musicbot.config import config
from musicbot.music.file import File, supported_formats
from musicbot.music.helpers import find_files, filecount

logger = logging.getLogger(__name__)

DEFAULT_MB_CONCURRENCY = 8
concurrency_options = [
    click.option(
        '--concurrency',
        help='Number of coroutines',
        default=DEFAULT_MB_CONCURRENCY,
        show_default=True
    )
]

DEFAULT_DRY = False
dry_option = [
    click.option(
        '--dry',
        help='Take no real action',
        default=DEFAULT_DRY,
        is_flag=True,
        show_default=True
    )
]

DEFAULT_YES = False
yes_option = [
    click.option(
        '-y', '--yes',
        help='Confirm action',
        default=DEFAULT_YES,
        is_flag=True,
    )
]

DEFAULT_SAVE = False
save_option = [
    click.option(
        '--save', '-s',
        help='Save to config file',
        default=DEFAULT_SAVE,
        is_flag=True,
        show_default=True
    )
]

DEFAULT_MB_OUTPUT = 'table'
output_option = [
    click.option(
        '--output',
        help='Output format',
        default=DEFAULT_MB_OUTPUT,
        show_default=True,
        type=click.Choice(['table', 'json'])
    )
]

DEFAULT_MB_PLAYLIST_OUTPUT = 'table'
playlist_output_option = [
    click.option(
        '--output',
        help='Output format',
        default=DEFAULT_MB_PLAYLIST_OUTPUT,
        show_default=True,
        type=click.Choice(['json', 'm3u', 'table'])
    )
]


def config_string(ctx: click.Context, param: Any, value: Any) -> Any:
    arg_value = value
    logger.info(f"{param.name} : try loading with value : {value}")

    config_value = config.configfile.get('musicbot', param.name, fallback=None)
    logger.info(f"{param.name} : try loading with config key : {config_value}")

    if arg_value:
        value = arg_value

    if config_value:
        if not value or value == param.default:
            value = config_value
        elif arg_value and arg_value != config_value:
            logger.warning(f"{param.name} : config value {config_value} is not sync with arg value {arg_value}")

    if not value and param.required:
        raise click.BadParameter(f'missing arg or config {param.name} in {config.config}', ctx, param.name, param.name)
    logger.info(f"{param.name} : final value {value}")
    ctx.params[param.name] = value
    return value


def config_list(ctx: click.Context, param: Any, value: Any) -> Any:
    arg_value = value
    logger.info(f"{param.name} : try loading with value : {value}")

    config_value = config.configfile.get('musicbot', param.name, fallback=None)
    list_value = tuple(mysplit(config_value, ',')) if config_value is not None else []
    logger.info(f"{param.name} : try loading with config key : {list_value}")

    if arg_value:
        value = arg_value

    if list_value:
        if not value or value == param.default:
            value = list_value
        elif arg_value and arg_value != list_value:
            logger.warning(f"{param.name} : config value {list_value} is not sync with arg value {arg_value}")

    if not value and param.required:
        raise click.BadParameter(f'missing arg or config {param.name} in {config.config}', ctx, param.name, param.name)
    logger.info(f"{param.name} : final value {value}")
    ctx.params[param.name] = value
    return value


@config.timeit
def genfiles(folders: Iterable[str]) -> Collection[File]:
    directories = [os.path.abspath(f) for f in folders]
    enabled = len(directories) and not config.quiet
    with enlighten.Manager(stream=sys.stderr, enabled=enabled) as manager:
        count = 0
        with manager.counter(total=len(directories), desc=f"Music counting {folders}") as pbar:
            for directory in directories:
                try:
                    pbar.desc = f"Music counting {directory}"
                    subcount = filecount(directory, supported_formats)
                    logger.info(f"{directory} : file count: {subcount}")
                    count += subcount
                finally:
                    pbar.update()
        files = []
        enabled = count and not config.quiet
        with manager.counter(total=count, desc="Music listing", enabled=enabled) as pbar:
            file_list = find_files(folders, supported_formats)
            music_files = list(file_list)
            for f in music_files:
                try:
                    m = File(f[1], f[0])
                    files.append(m)
                except OSError as e:
                    logger.error(e)
                finally:
                    pbar.update()
    return files


folders_argument = [
    click.argument(
        'folders',
        nargs=-1,
        callback=config_list,
        type=click.Path(exists=True, file_okay=False),
    )
]



import sys
from typing import Iterable, Dict, Optional, IO
from colorama import Fore  # type: ignore
from prettytable import PrettyTable  # type: ignore


def print_playlist(
    tracks: Iterable[Dict[str, str]],
    file: Optional[IO[str]] = None,
    current_title: Optional[str] = None,
    current_album: Optional[str] = None,
    current_artist: Optional[str] = None,
) -> None:
    if not tracks:
        return
    file = file if file is not None else sys.stdout
    pt = PrettyTable(["Title", "Album", "Artist"])
    for t in tracks:
        title = t.get('title', None)
        album = t.get('album', None)
        artist = t.get('artist', None)

        colored_title = (Fore.GREEN + title + Fore.RESET) if title == current_title else title
        colored_album = (Fore.GREEN + album + Fore.RESET) if album == current_album else album
        colored_artist = (Fore.GREEN + artist + Fore.RESET) if artist == current_artist else artist
        pt.add_row([colored_title, colored_album, colored_artist])
    print(pt, file=file)



import logging
import platform
import os
from typing import Any, Iterable, Dict
from prompt_toolkit import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.application import run_in_terminal, get_app
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit import HTML, print_formatted_text
from click_skeleton.helpers import seconds_to_human
from musicbot.playlist import print_playlist

logger = logging.getLogger(__name__)
logging.getLogger("vlc").setLevel(logging.NOTSET)


def play(tracks: Iterable[Dict[str, str]]) -> None:
    if not tracks:
        logger.warning('Empty playlist')
        return

    try:
        if platform.system() == 'Windows':
            os.add_dll_directory(r'C:\Program Files\VideoLAN\VLC')  # type: ignore
        import vlc  # type: ignore
        instance = vlc.Instance()
        if not instance:
            logger.critical('Unable to start VLC instance')
            return

        player = instance.media_list_player_new()
        if not player:
            logger.critical('Unable to create VLC player')
            return

        tracks_path = [track['path'] for track in tracks]
        media_list = instance.media_list_new(tracks_path)
        if not media_list:
            logger.critical('Unable to create VLC media list')
            return

        player.set_media_list(media_list)
        bindings = KeyBindings()

        def print_help() -> None:
            print_formatted_text(HTML('<violet>Bindings: q = quit | p = play | s = pause/continue | right = next song | left = previous song | l = playlist</violet>'))

        @bindings.add('p')
        def _play_binding(event: Any) -> None:  # pylint: disable=unused-argument
            def play() -> None:
                """Play song"""
                player.play()
            run_in_terminal(play)

        @bindings.add('q')
        def _quit_binding(event: Any) -> None:
            player.pause()
            event.app.exit()

        @bindings.add('s')
        def _pause_binding(event: Any) -> None:  # pylint: disable=unused-argument
            player.pause()

        @bindings.add('l')
        def _playlist_binding(event: Any) -> None:  # pylint: disable=unused-argument
            def playlist() -> None:
                """List songs"""
                media_player = player.get_media_player()
                media = media_player.get_media()
                media.parse()
                current_artist = media.get_meta(vlc.Meta.Artist)
                current_album = media.get_meta(vlc.Meta.Album)
                current_title = media.get_meta(vlc.Meta.Title)
                print_playlist(tracks, current_artist=current_artist, current_album=current_album, current_title=current_title)
            run_in_terminal(playlist)

        @bindings.add('right')
        def _next_binding(event: Any) -> None:  # pylint: disable=unused-argument
            player.next()

        @bindings.add('left')
        def _previous_binding(event: Any) -> None:  # pylint: disable=unused-argument
            player.previous()

        @bindings.add('h')
        def _help(event: Any) -> None:  # pylint: disable=unused-argument
            def _print_help() -> None:
                print_help()
            run_in_terminal(_print_help)
            get_app().invalidate()

        def bottom_toolbar() -> Any:
            media_player = player.get_media_player()
            media = media_player.get_media()
            media.parse()
            media_time = seconds_to_human(round(media_player.get_time() / 1000))
            media_length = seconds_to_human(round(media_player.get_length() / 1000))
            artist = media.get_meta(vlc.Meta.Artist)
            album = media.get_meta(vlc.Meta.Album)
            title = media.get_meta(vlc.Meta.Title)
            current = f'({media_time} / {media_length}) {artist} - {album} - {title}'
            get_app().invalidate()
            return HTML(f'Current song: {current}')

        print_help()
        player.play()

        root_container = HSplit(
            [Window(FormattedTextControl(lambda: bottom_toolbar, style='class:bottom-toolbar.text'), style='class:bottom-toolbar')]
        )
        layout = Layout(root_container)
        app: Any = Application(layout=layout, key_bindings=bindings)
        app.run()
    except NameError:
        version = vlc.libvlc_get_version()
        logger.critical(f"Your VLC version may be outdated: {version}")



import logging
import json
from prettytable import PrettyTable  # type: ignore
from click_skeleton import AdvancedGroup, add_options

from musicbot import helpers
from musicbot.cli import main_cli
from musicbot.user import User, register_options, auth_options, login_options
from musicbot.admin import Admin, admin_options
from musicbot.config import config

logger = logging.getLogger(__name__)


@main_cli.group('user', help='User management', cls=AdvancedGroup)
def user_cli():
    pass


@user_cli.command('list', help='List users (admin)')
@add_options(
    helpers.output_option +
    admin_options
)
def _list(graphql_admin, output, **kwargs):
    a = Admin(graphql=graphql_admin, **kwargs)
    users = a.users()
    if output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Email", "Firstname", "Lastname", "Created at", "Updated at"]
        for u in users:
            pt.add_row([u["email"], u["user"]["firstName"], u["user"]["lastName"], u["user"]["createdAt"], u["user"]["updatedAt"]])
        print(pt)
    elif output == 'json':
        print(json.dumps(users))


@user_cli.command(aliases=['new', 'add', 'create'], help='Register a new user')
@add_options(
    helpers.save_option +
    register_options
)
def register(save, **kwargs):
    u = User.register(**kwargs)
    if u.token and save:
        logger.info("saving user infos")
        config.configfile['musicbot']['email'] = u.email
        config.configfile['musicbot']['password'] = u.password
        config.configfile['musicbot']['token'] = u.token
        config.write()


@user_cli.command(aliases=['delete', 'remove'], help='Remove a user')
@add_options(auth_options)
def unregister(user):
    user.unregister()


@user_cli.command(aliases=['token'], help='Authenticate user')
@add_options(
    helpers.save_option +
    login_options
)
def login(user, save):
    print(user.token)
    if user.token and save:
        logger.info("saving user infos")
        config.configfile['musicbot']['token'] = user.token
        config.write()



import logging
from prettytable import PrettyTable  # type: ignore
from mutagen import MutagenError  # type: ignore
from click_skeleton import AdvancedGroup, add_options
from musicbot.cli import main_cli

from musicbot import helpers
from musicbot.music.file import File, keywords_argument, path_argument, folder_option, options, checks_options
from musicbot.music.fingerprint import acoustid_api_key_option

logger = logging.getLogger(__name__)


@main_cli.group('music', help='Music file', cls=AdvancedGroup, aliases=['file'])
def music_cli():
    pass


@music_cli.command(help='Convert flac music to mp3')
@add_options(
    path_argument +
    folder_option +
    helpers.dry_option
)
def flac2mp3(path, folder, dry):
    f = File(path)
    f.to_mp3(folder, dry)


@music_cli.command(help='Print music fingerprint')
@add_options(
    path_argument +
    acoustid_api_key_option
)
def fingerprint(path, acoustid_api_key):
    f = File(path)
    print(f.fingerprint(acoustid_api_key))


@music_cli.command(help='Print music tags')
@add_options(
    path_argument
)
def tags(path):
    f = File(path)
    logger.info(f.handle.tags.keys())
    print(f.as_dict())


@music_cli.command(aliases=['consistency'], help='Check music consistency')
@add_options(
    folder_option +
    path_argument +
    helpers.dry_option +
    checks_options
)
def inconsistencies(path, folder, fix, **kwargs):
    m = File(path, folder)
    pt = PrettyTable()
    pt.field_names = ["Folder", "Path", "Inconsistencies"]
    try:
        if fix:
            m.fix(**kwargs)
        if m.inconsistencies:
            pt.add_row([m.folder, m.path, ', '.join(m.inconsistencies)])
    except (OSError, MutagenError):
        pt.add_row([m.folder, m.path, "could not open file"])
    print(pt)


@music_cli.command(help='Set music title')
@add_options(
    path_argument +
    helpers.dry_option +
    options
)
def set_tags(path, title, artist, album, genre, keywords, rating, number, dry):
    f = File(path)
    if title:
        f.title = title
    if artist:
        f.artist = artist
    if album:
        f.album = album
    if genre:
        f.genre = genre
    if keywords:
        f.keywords = keywords
    if rating:
        f.rating = rating
    if number:
        f.number = number
    f.save(dry)


@music_cli.command(help='Add keywords to music')
@add_options(
    helpers.dry_option +
    path_argument +
    keywords_argument
)
def add_keywords(path, keywords, dry):
    f = File(path)
    f.add_keywords(keywords, dry)


@music_cli.command(help='Delete keywords to music')
@add_options(
    helpers.dry_option +
    path_argument +
    keywords_argument
)
def delete_keywords(path, keywords, dry):
    f = File(path)
    f.delete_keywords(keywords, dry)



import click
from click_option_group import optgroup, MutuallyExclusiveOptionGroup  # type: ignore
from click_skeleton import ExpandedPath
from musicbot.config import (
    VERBOSITIES,
    DEFAULT_QUIET,
    DEFAULT_CONFIG,
    DEFAULT_LOG,
    DEFAULT_VERBOSITY,
    DEFAULT_DEBUG,
    DEFAULT_INFO,
    DEFAULT_WARNING,
    DEFAULT_ERROR,
    DEFAULT_CRITICAL,
    DEFAULT_TIMINGS,

    MB_CONFIG,
    MB_LOG,
    MB_DEBUG,
    MB_INFO,
    MB_WARNING,
    MB_ERROR,
    MB_CRITICAL,
    MB_TIMINGS,
    MB_VERBOSITY,
    MB_QUIET,
)


config_option = [
    optgroup.option(
        '--config', '-c',
        help='Config file path',
        type=ExpandedPath(writable=True, dir_okay=False),
        envvar=MB_CONFIG,
        default=DEFAULT_CONFIG,
        show_default=True,
    )
]

log_option = [
    optgroup.option(
        '--log', '-l',
        help='Log file path',
        type=ExpandedPath(writable=True, dir_okay=False),
        envvar=MB_LOG,
        default=DEFAULT_LOG,
    )
]

debug_option = [
    optgroup.option(
        '--debug',
        help='Same as "--verbosity debug"',
        envvar=MB_DEBUG,
        default=DEFAULT_DEBUG,
        is_flag=True,
    )
]

info_option = [
    optgroup.option(
        '--info',
        help='Same as "--verbosity info"',
        envvar=MB_INFO,
        default=DEFAULT_INFO,
        is_flag=True,
    )
]

warning_option = [
    optgroup.option(
        '--warning',
        help='Same as "--verbosity warning"',
        envvar=MB_WARNING,
        default=DEFAULT_WARNING,
        is_flag=True,
    )
]

error_option = [
    optgroup.option(
        '--error',
        help='Same as "--verbosity error"',
        envvar=MB_ERROR,
        default=DEFAULT_ERROR,
        is_flag=True,
    )
]

critical_option = [
    optgroup.option(
        '--critical',
        help='Same as "--verbosity critical"',
        envvar=MB_CRITICAL,
        default=DEFAULT_CRITICAL,
        is_flag=True,
    )
]

timings_option = [
    optgroup.option(
        '--timings', '-t',
        help='Set verbosity to info and show execution timings',
        envvar=MB_TIMINGS,
        default=DEFAULT_TIMINGS,
        is_flag=True,
        show_default=True,
    )
]

verbosity_option = [
    optgroup.option(
        '--verbosity',
        help='Set verbosity level',
        envvar=MB_VERBOSITY,
        default=DEFAULT_VERBOSITY,
        type=click.Choice(VERBOSITIES.keys()),
        show_default=True,
    )
]

quiet_option = [
    optgroup.option(
        '--quiet', '-q',
        help='Disable progress bars',
        envvar=MB_QUIET,
        default=DEFAULT_QUIET,
        is_flag=True,
        show_default=True,
    )
]

options =\
    [optgroup.group('Global options')] +\
    config_option +\
    log_option +\
    quiet_option +\
    timings_option +\
    [optgroup('Verbosity', cls=MutuallyExclusiveOptionGroup)] +\
    debug_option +\
    info_option +\
    warning_option +\
    error_option +\
    critical_option +\
    verbosity_option



import logging
from watchdog.events import FileSystemEvent, PatternMatchingEventHandler  # type: ignore
from musicbot.music import file
from musicbot.user import User

logger = logging.getLogger(__name__)


class MusicWatcherHandler(PatternMatchingEventHandler):  # type: ignore
    def __init__(self, user: User) -> None:
        PatternMatchingEventHandler.__init__(
            self,
            patterns=['*.' + f for f in file.supported_formats],
            ignore_directories=True,
        )
        self.user = user

    def on_modified(self, event: FileSystemEvent) -> None:
        self.update_music(event.src_path)

    def on_created(self, event: FileSystemEvent) -> None:
        self.update_music(event.src_path)

    def on_deleted(self, event: FileSystemEvent) -> None:
        logger.debug(f'Deleting entry in DB for: {event.src_path} {event.event_type}')
        self.user.delete_music(event.src_path)

    def on_moved(self, event: FileSystemEvent) -> None:
        logger.debug(f'Moving entry in DB for: {event.src_path} {event.event_type}')
        self.user.delete_music(event.src_path)
        self.update_music(event.dest_path)

    def update_music(self, path: str) -> None:
        for folder in self.user.folders():
            if path.startswith(folder) and path.endswith(tuple(file.supported_formats)):
                logger.debug(f'Creating/modifying DB for: {path}')
                f = file.File(path, folder)
                self.user.upsert_music(f)
                return



import os
import logging
import click
import acoustid  # type: ignore
import youtube_dl  # type: ignore
import humanfriendly  # type: ignore
from click_skeleton import AdvancedGroup, add_options

from musicbot.cli import main_cli
from musicbot.music.file import File, path_argument
from musicbot.music.fingerprint import acoustid_api_key_option

logger = logging.getLogger(__name__)


@main_cli.group('youtube', help='Youtube tool', cls=AdvancedGroup)
def youtube_cli():
    pass


@youtube_cli.command(help='Search a youtube link with artist and title')
@click.argument('artist')
@click.argument('title')
def search(artist, title):
    '''Search a youtube link with artist and title'''
    ydl_opts = {
        'format': 'bestaudio',
        'ignoreerrors': True,
        'skip_download': True,
        'quiet': True,
        'no_warnings': True,
    }
    try:
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            infos = ydl.extract_info(f"ytsearch1:'{artist} {title}'", download=False)
            for entry in infos['entries']:
                print(entry['webpage_url'])
    except Exception as e:  # pylint: disable=broad-except
        logger.error(e)


@youtube_cli.command(help='Download a youtube link with artist and title')
@click.argument('artist')
@click.argument('title')
@click.option('--path', default=None)
def download(artist, title, path):
    try:
        if not path:
            path = f"{artist} - {title}.mp3"
        ydl_opts = {
            'format': 'bestaudio/best',
            'cachedir': False,
            'postprocessors': [{
                'key': 'FFmpegExtractAudio',
                'preferredcodec': 'mp3',
                'preferredquality': '192',
            }],
            'outtmpl': path,
        }
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            ydl.extract_info(f"ytsearch1:'{artist} {title}'", download=True)
    except youtube_dl.utils.DownloadError as e:
        logger.error(e)


@youtube_cli.command(help='Search a youtube link with artist and title')
@add_options(
    path_argument +
    acoustid_api_key_option
)
def find(path, acoustid_api_key):
    f = File(path)
    yt_path = f"{f.artist} - {f.title}.mp3"
    try:
        file_id = f.fingerprint(acoustid_api_key)
        print(f'Searching for artist {f.artist} and title {f.title} and duration {humanfriendly.format_timespan(f.duration)}')
        ydl_opts = {
            'format': 'bestaudio/best',
            'quiet': True,
            'cachedir': False,
            'no_warnings': True,
            'postprocessors': [{
                'key': 'FFmpegExtractAudio',
                'preferredcodec': 'mp3',
                'preferredquality': '192',
            }],
            'outtmpl': yt_path,
        }
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            infos = ydl.extract_info(f"ytsearch1:'{f.artist} {f.title}'", download=True)
            url = None
            for entry in infos['entries']:
                url = entry['webpage_url']
                break

            yt_ids = acoustid.match(acoustid_api_key, yt_path)
            yt_id = None
            for _, recording_id, _, _ in yt_ids:
                yt_id = recording_id
                break
            if file_id == yt_id:
                print(f'Found: fingerprint {file_id} | url {url}')
            else:
                print(f'Not exactly found: fingerprint file: {file_id} | yt: {yt_id} | url {url}')
                print(f'Based only on duration, maybe: {url}')
    except acoustid.WebServiceError as e:
        logger.error(e)
    except youtube_dl.utils.DownloadError as e:
        logger.error(e)
    finally:
        try:
            if yt_path:
                os.remove(yt_path)
        except OSError:
            logger.warning(f"File not found: {yt_path}")


@youtube_cli.command(help='Fingerprint a youtube video')
@click.argument('url')
@add_options(acoustid_api_key_option)
def fingerprint(url, acoustid_api_key):
    yt_path = "intermediate.mp3"
    ydl_opts = {
        'format': 'bestaudio/best',
        'quiet': True,
        'no_warnings': True,
        'postprocessors': [{
            'key': 'FFmpegExtractAudio',
            'preferredcodec': 'mp3',
            'preferredquality': '192',
        }],
        'outtmpl': yt_path,
    }
    try:
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            ydl.extract_info(url, download=True)
            yt_ids = acoustid.match(acoustid_api_key, yt_path)
            for _, recording_id, _, _ in yt_ids:
                print(recording_id)
                break
    except youtube_dl.utils.DownloadError as e:
        logger.error(e)
    except acoustid.WebServiceError as e:
        logger.error(e)



import logging
import sys
import json
import concurrent.futures as cf
import enlighten  # type: ignore
import mutagen  # type: ignore
from prettytable import PrettyTable  # type: ignore
from click_skeleton import AdvancedGroup, add_options

from musicbot import helpers
from musicbot.cli import main_cli
from musicbot.exceptions import MusicbotError
from musicbot.config import config
from musicbot.music.file import File, folder_option, checks_options, supported_formats
from musicbot.music.helpers import find_files

logger = logging.getLogger(__name__)


@main_cli.group('folder', help='Manage folders', cls=AdvancedGroup)
def folder_cli():
    pass


@folder_cli.command(help='Just list music files')
@add_options(helpers.folders_argument)
def find(folders):
    files = find_files(folders, supported_formats)
    for f in files:
        print(f[1])


@folder_cli.command(help='List tracks')
@add_options(
    helpers.folders_argument +
    helpers.output_option
)
def tracks(folders, output):
    tracks = helpers.genfiles(folders)
    if output == 'json':
        tracks_dict = [{'title': t.title, 'artist': t.artist, 'album': t.album} for t in tracks]
        print(json.dumps(tracks_dict))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Track", "Title", "Artist", "Album"]
        for t in tracks:
            pt.add_row([t.number, t.title, t.artist, t.album])
        print(pt)
    else:
        raise NotImplementedError


@folder_cli.command(help='Convert all files in folders to mp3')
@add_options(
    folder_option +
    helpers.folders_argument +
    helpers.concurrency_options +
    helpers.dry_option
)
def flac2mp3(folders, folder, concurrency, dry):
    flac_files = list(find_files(folders, ['flac']))
    if not flac_files:
        logger.warning(f"No flac files detected in {folders}")
        return

    enabled = not config.quiet
    with enlighten.Manager(stream=sys.stderr, enabled=enabled) as manager:
        with manager.counter(total=len(flac_files), desc="converting musics") as pbar:
            with cf.ThreadPoolExecutor(max_workers=concurrency) as executor:
                def convert(flac_path):
                    try:
                        f = File(flac_path)
                        f.to_mp3(folder, dry)
                    except MusicbotError as e:
                        logger.error(e)
                    finally:
                        pbar.update()
                executor.shutdown = lambda wait: None
                futures = [executor.submit(convert, flac_path[1]) for flac_path in flac_files]
                cf.wait(futures)


@folder_cli.command(aliases=['consistency'], help='Check music files consistency')
@add_options(
    helpers.folders_argument +
    helpers.dry_option +
    checks_options
)
def inconsistencies(folders, fix, **kwargs):
    musics = helpers.genfiles(folders)
    pt = PrettyTable()
    pt.field_names = ["Folder", "Path", "Inconsistencies"]
    for m in musics:
        try:
            if fix:
                m.fix(**kwargs)
            if m.inconsistencies:
                pt.add_row([m.folder, m.path, ', '.join(m.inconsistencies)])
        except (OSError, mutagen.MutagenError):
            pt.add_row([m.folder, m.path, "could not open file"])
    print(pt)



import logging
import json
from typing import Any, Optional
import requests
from musicbot.exceptions import FailedRequest, FailedAuthentication

logger = logging.getLogger(__name__)


class GraphQL:
    def __init__(self, graphql: str, authorization: Optional[str] = None) -> None:
        self.graphql = graphql
        self._headers = {'Authorization': authorization} if authorization else {}

    def post(self, query: str) -> Any:
        response = requests.post(
            self.graphql,
            headers=self._headers,
            json={'query': query},
        )
        logger.debug(response)
        if response.status_code == 401:
            raise FailedAuthentication(f"Authentication failed: {response.text} | {self._headers}")

        try:
            json_response = response.json()
        except json.JSONDecodeError as e:
            raise FailedRequest(f"Query failed : {query} | {self._headers}") from e

        logger.debug(json_response)
        if 'errors' in json_response and json_response['errors']:
            errors = [e['message'] for e in json_response['errors']]
            raise FailedRequest(f"Query failed: {query} | {errors} | {self._headers}")
        return json_response



import logging
import os
import itertools
from typing import List, Any
import spotipy  # type: ignore
import click
from click_option_group import optgroup  # type: ignore
from click_skeleton import ExpandedPath
from musicbot.config import config

logger = logging.getLogger(__name__)


def config_string_spotify(ctx, param, arg_value):
    name = param.name
    value = None

    config_value = config.configfile.get('spotify', name, fallback=None)
    if config_value:
        value = config_value
        logger.info(f"{name} : config key loaded : {config_value}")

    name_upper = name.upper().replace('-', '_')
    env_name = 'MB_SPOTIFY_' + name_upper
    env_value = os.getenv(env_name, None)
    if env_value:
        value = env_value
        logger.info(f"{name} : env key {env_name} loaded : {env_value}")

    default_name = 'DEFAULT_' + name_upper
    default_value = globals().get(default_name, None)
    logger.info(f'{name} : default_name: {default_name} default_value: {default_value}')

    if arg_value and arg_value != default_value:
        logger.info(f"{name} : arg value loaded : {arg_value}")
        value = arg_value

    if not value:
        logger.info(f"{name} : default arg value loaded {default_name} : {arg_value}")
        value = default_value

    if env_value and arg_value and env_value != arg_value:
        logger.info(f"{name} : env value {env_value} is not sync with arg value {arg_value}")
    if env_value and config_value and env_value != config_value:
        logger.info(f"{name} : config value {config_value} is not sync with env value {env_value}")
    if arg_value and config_value and arg_value != config_value:
        logger.info(f"{name} : config value {config_value} is not sync with arg value {arg_value}")

    if not value and param.required:
        raise click.BadParameter(f'missing arg {param.name} or env {env_name} or {param.name} in [spotify] section of {config.config}', ctx, param.name, param.name)
    logger.info(f"{name} : final value {value}")
    return value


def sane_spotify(ctx, param, value):  # pylint: disable=unused-argument
    kwargs = {}
    for field in ('username', 'client_id', 'client_secret', 'cache_path', 'redirect_uri', 'scopes'):
        kwargs[field] = ctx.params[field]
        ctx.params.pop(field)

    ctx.params['spotify'] = Spotify(**kwargs)
    return ctx.params['spotify']


DEFAULT_CACHE_PATH = '~/.spotify_cache'
cache_path_option = [
    optgroup.option(
        '--cache-path',
        help='Spotify cache path',
        is_eager=True,
        type=ExpandedPath(writable=True, readable=True, dir_okay=False),
        default=DEFAULT_CACHE_PATH,
        callback=config_string_spotify,
    )
]

DEFAULT_SCOPES = 'user-library-read,user-follow-read,user-top-read,playlist-read-private,user-modify-playback-state,user-read-currently-playing,user-read-playback-state'
scopes_option = [
    optgroup.option(
        '--scopes',
        help='Spotify scopes',
        is_eager=True,
        default=DEFAULT_SCOPES,
        callback=config_string_spotify,
    )
]

DEFAULT_REDIRECT_URI = 'http://localhost:8888/spotify/callback'
redirect_uri_option = [
    optgroup.option(
        '--redirect-uri',
        help='Spotify redirect URI',
        is_eager=True,
        default=DEFAULT_REDIRECT_URI,
        callback=config_string_spotify,
    )
]

username_option = [
    optgroup.option(
        '--username',
        help='Spotify username',
        is_eager=True,
        callback=config_string_spotify,
    )
]
client_id_option = [
    optgroup.option(
        '--client-id',
        help='Spotify client ID',
        is_eager=True,
        callback=config_string_spotify,
    )
]
client_secret_option = [
    optgroup.option(
        '--client-secret',
        help='Spotify client secret',
        is_eager=True,
        callback=config_string_spotify,
    )
]

token_option = [
    optgroup.option(
        '--token',
        help='Spotify token',
        expose_value=False,
        callback=sane_spotify,
    )
]

spotify_options =\
    [optgroup.group('Spotify options')] +\
    username_option +\
    client_id_option +\
    client_secret_option +\
    token_option +\
    cache_path_option +\
    scopes_option +\
    redirect_uri_option


class Spotify:
    def __init__(self, username: str, client_id: str, client_secret: str, cache_path: str, scopes: str, redirect_uri: str):
        self.scopes = scopes
        self.redirect_uri = redirect_uri
        self.cache_path = cache_path
        self.username = username
        self.client_id = client_id
        self.client_secret = client_secret

        self.auth_manager = spotipy.oauth2.SpotifyOAuth(
            client_id=self.client_id,
            client_secret=self.client_secret,
            username=self.username,
            cache_path=self.cache_path,
            redirect_uri=self.redirect_uri,
            scope=self.scopes,
        )
        self.sp = spotipy.Spotify(auth_manager=self.auth_manager)
        # self.token = spotipy.util.prompt_for_user_token(
        #     client_id=self.client_id,
        #     client_secret=self.client_secret,
        #     username=self.username,
        #     cache_path=self.cache_path,
        #     redirect_uri=self.redirect_uri,
        #     scope=self.scopes,
        # )
        # self.sp = spotipy.Spotify(auth=self.token)

    @property
    def token(self):
        return self.auth_manager.get_cached_token()

    def __repr__(self):
        return f'token: {self.token} | username: {self.username} | client_id: {self.client_id} | client_secret: {self.client_secret} | cache: {self.cache_path} | redirect uri: {self.redirect_uri} | scopes: {self.scopes}'

    def tracks(self) -> List[Any]:
        offset = 0
        limit = 50
        objects = []
        while True:
            new_objects = self.sp.current_user_saved_tracks(limit=limit, offset=offset)
            objects.append(new_objects['items'])
            length = len(new_objects['items'])
            logger.info(f"chunk size: {length}")
            offset += len(new_objects['items'])
            if len(new_objects['items']) < limit:
                break
        return list(itertools.chain(*objects))

    def playlists(self) -> List[Any]:
        offset = 0
        limit = 50
        objects = []
        while True:
            new_objects = self.sp.current_user_playlists(limit=limit, offset=offset)
            length = len(new_objects['items'])
            objects.append(new_objects['items'])
            logger.info(f"chunk size: {length}")
            offset += length
            if length < limit:
                break
        logger.debug(f"length: {offset}")
        return list(itertools.chain(*objects))

    def playlist(self, name: str) -> List[Any]:
        playlists = self.playlists()
        for p in playlists:
            if p['name'] == name:
                tracks = []
                results = self.sp.playlist(p['id'], fields="tracks,next")
                new_tracks = results['tracks']
                tracks.append(new_tracks['items'])
                while new_tracks['next']:
                    new_tracks = self.sp.next(new_tracks)
                    tracks.append(new_tracks['tracks']['items'])
                return list(itertools.chain(*tracks))
        return []



class MusicbotError(Exception):
    pass


class MusicbotConfigError(MusicbotError):
    pass


class FailedRegistration(MusicbotError):
    pass


class FailedAuthentication(MusicbotError):
    pass


class FailedRequest(MusicbotError):
    pass



import logging
import io
import sys
import shutil
import os
import codecs
import json
import datetime
import textwrap
import click
import enlighten  # type: ignore
import attr
import mutagen  # type: ignore
from prettytable import PrettyTable  # type: ignore
from click_skeleton import AdvancedGroup, add_options

from musicbot import helpers, user
from musicbot.cli import main_cli
from musicbot.player import play
from musicbot.playlist import print_playlist
from musicbot.config import config
from musicbot.music import mfilter
from musicbot.music.file import File, checks_options, folder_argument
from musicbot.music.helpers import bytes_to_human, all_files, empty_dirs, except_directories


logger = logging.getLogger(__name__)


@main_cli.group('local', help='''Local music management''', cls=AdvancedGroup)
def local_cli():
    pass


@local_cli.command(help='Count musics')
@add_options(user.auth_options)
def count(user):
    print(user.count())


@local_cli.command(help='Raw query')
@click.argument('query')
@add_options(user.auth_options)
def execute(user, query):
    print(json.dumps(user.post(query)['data']))


@local_cli.command(help='Load default filters')
@add_options(user.auth_options)
def load_filters(user):
    user.load_default_filters()


@local_cli.command(help='List filters')
@add_options(
    helpers.output_option +
    user.auth_options
)
def filters(user, output):
    if output == 'json':
        print(json.dumps(user.filters()))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Name", "Keywords", "No keywords", "Min rating", "Max rating"]
        for f in user.filters():
            pt.add_row([f['name'], f['keywords'], f['noKeywords'], f['minRating'], f['maxRating']])
        print(pt)


@local_cli.command('filter', help='Print a filter')
@add_options(
    helpers.output_option +
    user.auth_options
)
@click.argument('name')
def _filter(user, name, output):
    f = user.filter(name)
    if output == 'json':
        print(json.dumps(f))
    elif output == 'table':
        print(f)


@local_cli.command(aliases=['stat'], help='Generate some stats for music collection with filters')
@add_options(
    helpers.output_option +
    user.auth_options +
    mfilter.mfilter_options
)
def stats(user, output, music_filter):
    stats = user.do_stat(music_filter)
    if output == 'json':
        print(json.dumps(stats))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Stat", "Value"]
        pt.add_row(["Music", stats['musics']])
        pt.add_row(["Artist", stats['artists']])
        pt.add_row(["Album", stats['albums']])
        pt.add_row(["Genre", stats['genres']])
        pt.add_row(["Keywords", stats['keywords']])
        pt.add_row(["Size", bytes_to_human(int(stats['size']))])
        pt.add_row(["Total duration", datetime.timedelta(seconds=int(stats['duration']))])
        print(pt)


@local_cli.command(help='List folders')
@add_options(
    helpers.output_option +
    user.auth_options
)
def folders(user, output):
    _folders = user.folders()
    if output == 'json':
        print(json.dumps(_folders))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Folder"]
        for f in _folders:
            pt.add_row([f])
        print(pt)


@local_cli.command(help='Load musics')
@add_options(
    helpers.folders_argument +
    user.auth_options
)
def scan(user, folders):
    if not folders:
        folders = user.folders()
    files = helpers.genfiles(folders)
    user.bulk_insert(files)


@local_cli.command(help='Watch files changes in folders')
@add_options(user.auth_options)
def watch(user):
    user.watch()


@local_cli.command(help='Clean all musics')
@add_options(
    user.auth_options +
    helpers.yes_option
)
def clean(user, yes):
    if yes or click.confirm("Are you sure to delete all musics from DB?"):
        user.clean_musics()


@local_cli.command(help='Clean and load musics')
@add_options(
    helpers.folders_argument +
    user.auth_options
)
def rescan(user, folders):
    if not folders:
        folders = user.folders()
    files = helpers.genfiles(folders)
    user.clean_musics()
    user.bulk_insert(files)


@local_cli.command(help='Copy selected musics with filters to destination folder')
@add_options(
    helpers.dry_option +
    user.auth_options +
    mfilter.mfilter_options
)
@click.argument('destination')
def sync(user, dry, destination, music_filter):
    logger.info(f'Destination: {destination}')
    musics = user.do_filter(music_filter)

    files = list(all_files(destination))
    logger.info(f"Files : {len(files)}")
    if not files:
        logger.warning("No files found in destination")

    destinations = {f[len(destination) + 1:]: f for f in files}
    logger.info(f"Destinations : {len(destinations)}")
    sources = {m['path'][len(m['folder']) + 1:]: m['path'] for m in musics}
    logger.info(f"Sources : {len(sources)}")
    to_delete = set(destinations.keys()) - set(sources.keys())
    with enlighten.Manager(stream=sys.stderr) as manager:
        logger.info(f"To delete: {len(to_delete)}")
        enabled = to_delete and not config.quiet
        with manager.counter(total=len(to_delete), enabled=enabled) as pbar:
            for d in to_delete:
                try:
                    pbar.desc = f"Deleting musics and playlists: {os.path.basename(destinations[d])}"
                    if not dry:
                        try:
                            logger.info(f"Deleting {destinations[d]}")
                            os.remove(destinations[d])
                        except OSError as e:
                            logger.error(e)
                    else:
                        logger.info(f"[DRY-RUN] False Deleting {destinations[d]}")
                finally:
                    pbar.update()

        to_copy = set(sources.keys()) - set(destinations.keys())
        logger.info(f"To copy: {len(to_copy)}")
        enabled = to_copy and not config.quiet
        with manager.counter(total=len(to_copy), enabled=enabled) as pbar:
            for c in sorted(to_copy):
                final_destination = os.path.join(destination, c)
                try:
                    pbar.desc = f'Copying {os.path.basename(sources[c])} to {destination}'
                    if not dry:
                        logger.info(f"Copying {sources[c]} to {final_destination}")
                        os.makedirs(os.path.dirname(final_destination), exist_ok=True)
                        shutil.copyfile(sources[c], final_destination)
                    else:
                        logger.info(f"[DRY-RUN] False Copying {sources[c]} to {final_destination}")
                except KeyboardInterrupt:
                    logger.debug(f"Cleanup {final_destination}")
                    try:
                        os.remove(final_destination)
                    except OSError:
                        pass
                    raise
                finally:
                    pbar.update()

    for d in empty_dirs(destination):
        if any(e in d for e in except_directories):
            logger.debug(f"Invalid path {d}")
            continue
        if not dry:
            shutil.rmtree(d)
        logger.info(f"[DRY-RUN] Removing empty dir {d}")


@local_cli.command(help='Generate a new playlist')
@add_options(
    helpers.dry_option +
    helpers.playlist_output_option +
    user.auth_options +
    mfilter.mfilter_options
)
@click.argument('path', type=click.File('w'), default='-')
def playlist(user, output, path, dry, music_filter):
    if output == 'm3u':
        p = user.playlist(music_filter)
        if not dry:
            print(p, file=path)
        else:
            logger.info(f'DRY RUN: Writing playlist to {path} with content:\n{p}')
    elif output == 'json':
        tracks = user.do_filter(music_filter)
        print(json.dumps(tracks), file=path)
    elif output == 'table':
        tracks = user.do_filter(music_filter)
        print_playlist(tracks, path)


@local_cli.command(help='Generate bests playlists with some rules')
@click.option('--prefix', envvar='MB_PREFIX', help="Append prefix before each path (implies relative)", default='')
@click.option('--suffix', envvar='MB_SUFFIX', help="Append this suffix to playlist name", default='')
@add_options(
    folder_argument +
    helpers.dry_option +
    user.auth_options +
    mfilter.mfilter_options
)
def bests(user, dry, folder, prefix, suffix, music_filter):
    if prefix:
        music_filter = attr.evolve(music_filter, relative=True)
        if not prefix.endswith('/'):
            prefix += '/'
    playlists = user.bests(music_filter)
    enabled = playlists and not config.quiet
    with enlighten.Manager(stream=sys.stderr, enabled=enabled) as manager:
        with manager.counter(total=len(playlists)) as pbar:
            for p in playlists:
                try:
                    playlist_filepath = os.path.join(folder, p['name'] + suffix + '.m3u')
                    pbar.desc = f"Best playlist {prefix} {suffix}: {os.path.basename(playlist_filepath)}"
                    content = textwrap.indent(p['content'], prefix, lambda line: line != '#EXTM3U\n')
                    if not dry:
                        try:
                            with codecs.open(playlist_filepath, 'w', "utf-8-sig") as playlist_file:
                                logger.debug(f'Writing playlist to {playlist_filepath} with content:\n{content}')
                                playlist_file.write(content)
                        except (OSError, LookupError, ValueError, UnicodeError) as e:
                            logger.warning(f'Unable to write playlist to {playlist_filepath} because of {e}')
                    else:
                        logger.info(f'DRY RUN: Writing playlist to {playlist_filepath} with content:\n{content}')
                finally:
                    pbar.update()


@local_cli.command(aliases=['play'], help='Music player')
@add_options(
    user.auth_options +
    mfilter.mfilter_options
)
def player(user, music_filter):
    try:
        tracks = user.do_filter(music_filter)
        play(tracks)
    except io.UnsupportedOperation:
        logger.critical('Unable to load UI')


@local_cli.command(aliases=['consistency'], help='Check music consistency')
@add_options(
    checks_options +
    helpers.dry_option +
    user.auth_options +
    mfilter.mfilter_options
)
def inconsistencies(user, dry, fix, checks, music_filter):
    tracks = user.do_filter(music_filter)
    pt = PrettyTable()
    pt.field_names = ["Folder", "Path", "Inconsistencies"]
    for t in tracks:
        try:
            m = File(t['path'], t['folder'])
            if fix:
                m.fix(dry=dry, checks=checks)
            if m.inconsistencies:
                pt.add_row([m.folder, m.path, ', '.join(m.inconsistencies)])
        except (OSError, mutagen.MutagenError):
            pt.add_row([t['folder'], t['path'], "could not open file"])
    print(pt)



import logging
import time
import base64
import json
import functools
import sys
from typing import Any, Collection, Optional
import click
import enlighten  # type: ignore
from click_option_group import optgroup  # type: ignore
from watchdog.observers import Observer  # type: ignore
from musicbot.graphql import GraphQL
from musicbot.config import config
from musicbot.helpers import config_string
from musicbot.exceptions import MusicbotError, FailedAuthentication, FailedRegistration
from musicbot.music import file, mfilter

logger = logging.getLogger(__name__)


DEFAULT_EMAIL = None
email_option = [
    optgroup.option(
        '--email', '-e',
        help='User email',
        default=DEFAULT_EMAIL,
        is_eager=True,
        callback=config_string,
    )
]

DEFAULT_PASSWORD = None
password_option = [
    optgroup.option(
        '--password', '-p',
        help='User password',
        default=DEFAULT_PASSWORD,
        is_eager=True,
        callback=config_string,
    )
]

DEFAULT_FIRST_NAME = None
first_name_option = [
    optgroup.option(
        '--first-name',
        help='User first name',
        default=DEFAULT_FIRST_NAME,
        is_eager=True,
        callback=config_string,
        show_default=True,
    )
]

DEFAULT_LAST_NAME = None
last_name_option = [
    optgroup.option(
        '--last-name',
        help='User last name',
        default=DEFAULT_FIRST_NAME,
        is_eager=True,
        callback=config_string,
        show_default=True,
    )
]

DEFAULT_TOKEN = None
token_option = [
    optgroup.option(
        '--token', '-t',
        help='User token',
    )
]

DEFAULT_GRAPHQL = 'http://127.0.0.1:5000/graphql'
graphql_option = [
    optgroup.option(
        '--graphql',
        help='GraphQL endpoint',
        default=DEFAULT_GRAPHQL,
        is_eager=True,
        callback=config_string,
        show_default=True,
    )
]


class User(GraphQL):
    def __init__(self, graphql: str, email: Optional[str] = None, password: Optional[str] = None, token: Optional[str] = None) -> None:
        self.authenticated = False

        if token:
            self.token = token
            logger.debug(f"using token : {self.token}")
        elif email and password:
            query = f"""
            mutation
            {{
                authenticate(input: {{email: "{email}", password: "{password}"}})
                {{
                    jwtToken
                }}
            }}"""
            response = None
            try:
                graphql_auth = GraphQL(graphql=graphql)
                response = graphql_auth.post(query)
                self.token = response['data']['authenticate']['jwtToken']
                self.email = email
                self.password = password
            except MusicbotError as e:
                raise FailedAuthentication(f"Authentication failed for email {email}") from e
            except KeyError as e:
                raise FailedAuthentication(f"Invalid response received : {response}") from e
        else:
            raise FailedAuthentication("No credentials or token provided")
        GraphQL.__init__(self, graphql, authorization=f"Bearer {self.token}")
        self.authenticated = True

    @config.timeit
    def load_default_filters(self) -> Any:
        query = """
        mutation
        {
            default:           createFilter(input: {filter: {name: "default"                                                                           }}){clientMutationId}
            no_artist_set:     createFilter(input: {filter: {name: "no artist set",     artists:    ""                                                 }}){clientMutationId}
            no_album_set:      createFilter(input: {filter: {name: "no album set",      albums:     ""                                                 }}){clientMutationId}
            no_title_set:      createFilter(input: {filter: {name: "no title set",      titles:     ""                                                 }}){clientMutationId}
            no_genre_set:      createFilter(input: {filter: {name: "no genre set",      genres:     ""                                                 }}){clientMutationId}
            youtube_not_found: createFilter(input: {filter: {name: "youtube not found", youtubes:   ["not found"]                                      }}){clientMutationId}
            spotify_not_found: createFilter(input: {filter: {name: "spotify not found", spotifys:   ["not found"]                                      }}){clientMutationId}
            no_youtube_links:  createFilter(input: {filter: {name: "no youtube links",  youtubes:   []                                                 }}){clientMutationId}
            no_spotify_links:  createFilter(input: {filter: {name: "no spotify links",  spotifys:   ["not found"]                                      }}){clientMutationId}
            no_rating:         createFilter(input: {filter: {name: "no rating",         minRating:  0.0, maxRating: 0.0                                }}){clientMutationId}
            bests_40:          createFilter(input: {filter: {name: "best (4.0+)",       minRating:  4.0, noKeywords: ["cutoff", "bad", "demo", "intro"]}}){clientMutationId}
            bests_45:          createFilter(input: {filter: {name: "best (4.5+)",       minRating:  4.5, noKeywords: ["cutoff", "bad", "demo", "intro"]}}){clientMutationId}
            bests_50:          createFilter(input: {filter: {name: "best (5.0+)",       minRating:  5.0, noKeywords: ["cutoff", "bad", "demo", "intro"]}}){clientMutationId}
            no_live:           createFilter(input: {filter: {name: "no live",           noKeywords: ["live"]                                           }}){clientMutationId}
            only_live:         createFilter(input: {filter: {name: "only live",         keywords:   ["live"]                                           }}){clientMutationId}
        }"""
        return self.post(query)

    @config.timeit
    def playlist(self, mf: Optional[mfilter.Filter] = None) -> Any:
        mf = mf if mf is not None else mfilter.Filter()
        query = f"""
        {{
            playlist({mf.to_graphql()})
        }}"""
        return self.post(query)['data']['playlist']

    @config.timeit
    def bests(self, mf: Optional[mfilter.Filter] = None) -> Any:
        mf = mf if mf is not None else mfilter.Filter()
        query = f"""
        {{
            bests({mf.to_graphql()})
            {{
                nodes
                {{
                    name,
                    content
                }}
            }}
        }}"""
        return self.post(query)['data']['bests']['nodes']

    @config.timeit
    def do_filter(self, mf: Optional[mfilter.Filter] = None) -> Any:
        mf = mf if mf is not None else mfilter.Filter()
        if mf.name:
            kwargs = self.filter(mf.name)
            print(kwargs)
            mf = mfilter.Filter(**kwargs)

        query = f"""
        {{
            doFilter({mf.to_graphql()})
            {{
                nodes
                {{
                    title,
                    album,
                    genre,
                    artist,
                    folder,
                    youtube,
                    spotify,
                    number,
                    path,
                    rating,
                    duration,
                    size,
                    keywords
                }}
            }}
        }}"""
        return self.post(query)['data']['doFilter']['nodes']

    @config.timeit
    def do_stat(self, mf: Optional[mfilter.Filter] = None) -> Any:
        mf = mf if mf is not None else mfilter.Filter()
        query = f"""
        {{
            doStat({mf.to_graphql()})
            {{
              musics,
              artists,
              albums,
              genres,
              keywords,
              size,
              duration
            }}
        }}"""
        return self.post(query)['data']['doStat']

    @config.timeit
    def upsert_music(self, music: file.File) -> Any:
        query = f"""
        mutation
        {{
            upsertMusic(input: {{{music.to_graphql()}}})
            {{
                clientMutationId
            }}
        }}"""
        return self.post(query)

    @config.timeit
    def bulk_insert(self, musics: Collection[file.File]) -> Any:
        if not musics:
            logger.info("no musics to insert")
            return None
        if config.debug:
            with enlighten.Manager(stream=sys.stderr) as manager:
                with manager.counter(total=len(musics), desc="inserting music one by one") as pbar:
                    for music in musics:
                        logger.debug(f"inserting {music}")
                        self.upsert_music(music)
                        pbar.update()
            return None

        j = json.dumps([m.as_dict() for m in musics])
        b64 = j.encode('utf-8')
        data = base64.b64encode(b64)
        query = f'''
        mutation
        {{
            bulkInsert(input: {{data: "{data.decode()}"}})
            {{
                clientMutationId
            }}
        }}'''
        return self.post(query)

    @functools.lru_cache(maxsize=None)
    @config.timeit
    def folders(self) -> Any:
        query = """
        {
            foldersList
        }"""
        return self.post(query)['data']['foldersList']

    @functools.lru_cache(maxsize=None)
    @config.timeit
    def count(self) -> int:
        query = '''
        {
            rawMusics
            {
                totalCount
            }
        }
        '''
        return int(self.post(query)['data']['rawMusics']['totalCount'])

    @functools.lru_cache(maxsize=None)
    @config.timeit
    def artists(self) -> Any:
        query = """
        {
          artistsTreeList {
            name
            albums {
              name
              musics {
                folder
                name
                path
              }
            }
          }
        }"""
        return self.post(query)['data']['artistsTreeList']

    @functools.lru_cache(maxsize=None)
    @config.timeit
    def genres(self) -> Any:
        query = """
        {
          genresTreeList {
            name
          }
        }"""
        return self.post(query)['data']['genresTreeList']

    @functools.lru_cache(maxsize=None)
    @config.timeit
    def filter(self, name: str) -> Any:
        default_filter = mfilter.Filter()
        filter_members = ','.join(default_filter.as_dict().keys())
        query = f"""
        {{
            filtersList(filter: {{name: {{equalTo: "{name}"}}}})
            {{
                name,
                {filter_members}
            }}
        }}"""
        return self.post(query)['data']['filtersList'][0]

    @functools.lru_cache(maxsize=None)
    @config.timeit
    def filters(self) -> Any:
        default_filter = mfilter.Filter()
        filter_members = ','.join(default_filter.as_dict().keys())
        query = f"""
        {{
            filtersList
            {{
                name,
                {filter_members}
            }}
        }}"""
        return self.post(query)['data']['filtersList']

    def watch(self) -> None:
        from .watcher import MusicWatcherHandler
        logger.info(f'Watching: {self.folders()}')
        event_handler = MusicWatcherHandler(user=self)
        observer = Observer()
        for f in self.folders():
            observer.schedule(event_handler, f, recursive=True)
        observer.start()
        try:
            while True:
                time.sleep(50)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()

    @classmethod
    @config.timeit
    def register(cls, graphql: str, email: str, password: str, first_name: str, last_name: str) -> "User":
        query = f"""
        mutation
        {{
            registerUser(input: {{firstName: "{first_name}", lastName: "{last_name}", email: "{email}", password: "{password}"}})
            {{
                clientMutationId
            }}
        }}"""

        try:
            graphql_register = GraphQL(graphql=graphql)
            graphql_register.post(query)
        except MusicbotError as e:
            raise FailedRegistration(f"Registration failed for {first_name} | {last_name} | {email} | {password}") from e
        return User(
            graphql=graphql,
            email=email,
            password=password
        )

    @config.timeit
    def unregister(self) -> Any:
        query = """
        mutation
        {
            unregisterUser(input: {})
            {
                clientMutationId
            }
        }"""
        try:
            result = self.post(query)
            self.authenticated = False
            return result
        except MusicbotError as e:
            raise FailedAuthentication(f"Cannot delete user {self.email}") from e

    @config.timeit
    def delete_music(self, path: str) -> Any:
        query = f"""
        mutation
        {{
            deleteMusic(input: {{path: "{path}"}})
            {{
                clientMutationId
            }}
        }}"""
        return self.post(query)

    @config.timeit
    def clean_musics(self) -> Any:
        query = """
        mutation
        {
            deleteAllMusic(input: {})
            {
                clientMutationId
            }
        }"""
        return self.post(query)


def sane_user(ctx: click.Context, param: Any, value: Any) -> User:  # pylint: disable=unused-argument
    kwargs = {}
    for field in ('token', 'email', 'password', 'graphql'):
        kwargs[field] = ctx.params.get(field, None)
        ctx.params.pop(field, None)
    user = User(**kwargs)
    ctx.params['user'] = user
    return user


DEFAULT_FILTER = None
user_option = [
    optgroup.option(
        '--user',
        help='Music Filter',
        expose_value=False,
        callback=sane_user,
        hidden=True,
    )
]

register_options =\
    [optgroup.group('Register options')] +\
    graphql_option +\
    email_option +\
    password_option +\
    first_name_option +\
    last_name_option

login_options =\
    [optgroup.group('User options')] +\
    graphql_option +\
    email_option +\
    password_option +\
    user_option

auth_options =\
    [optgroup.group('Auth options')] +\
    graphql_option +\
    email_option +\
    password_option +\
    token_option +\
    user_option



#!/usr/bin/env python3
'''Main module, import commands and start CLI'''
import logging
import sys
import requests
import spotipy  # type: ignore
import mutagen  # type: ignore

from click_skeleton import backtrace, version_checker, helpers
from musicbot.config import config as config_obj
from musicbot.cli import main_cli, PROG_NAME, __version__
from musicbot import exceptions

logger = logging.getLogger(__name__)

backtrace.hook(
    reverse=False,
    align=True,
    strip_path=False,
    enable_on_envvar_only=False,
    on_tty=False,
    conservative=False,
)


def main(**kwargs):
    helpers.raise_limits()
    from musicbot import commands
    logger.debug(f"Imported commands: {commands.modules}")
    version_check = version_checker.VersionCheckerThread(
        prog_name=PROG_NAME,
        current_version=__version__,
        autostart=config_obj.check_version,
    )
    try:
        exit_code = main_cli.main(**kwargs)
        version_check.print()
        return exit_code
    except (mutagen.MutagenError, exceptions.MusicbotError, spotipy.client.SpotifyException, requests.exceptions.ConnectionError) as e:
        if config_obj.debug:
            logger.exception(e)
        else:
            raise


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['musicbot', 'musicbot.commands', 'musicbot.music']

package_data = \
{'': ['*'], 'musicbot': ['schema/*']}

install_requires = \
['attrs>=20.1.0,<21.0.0',
 'click-skeleton>=0.7.1,<0.8.0',
 'colorlog>=4.1.0,<5.0.0',
 'enlighten>=1.6.1,<2.0.0',
 'graphql-py>=0.7.1,<0.8.0',
 'humanfriendly>=8.1,<9.0',
 'jellyfish>=0.8.2,<0.9.0',
 'logging-tree>=1.8,<2.0',
 'mutagen>=1.44.0,<2.0.0',
 'prettytable>=0.7.2,<0.8.0',
 'prompt_toolkit>=3.0.2,<4.0.0',
 'pyacoustid>=1.1,<2.0',
 'pydub>=0.24.0,<0.25.0',
 'pyrsistent==0.17.2',
 'python-slugify>=4.0.0,<5.0.0',
 'python-vlc>=3.0,<4.0',
 'requests>=2.24.0,<3.0.0',
 'spotipy>=2.10.0,<3.0.0',
 'watchdog>=0.10.0,<0.11.0',
 'youtube-dl>=2020.1.24,<2021.0.0']

entry_points = \
{'console_scripts': ['musicbot = musicbot.main:main']}

setup_kwargs = {
    'name': 'musicbot',
    'version': '0.7.4',
    'description': 'Music swiss army knife',
    'long_description': '========\nMusicBot\n========\n+---------------+-----------------+\n|     Tools     |      Result     |\n+===============+=================+\n|     Codacy    |    |codacy|     |\n+---------------+-----------------+\n|     Coverage  |   |coverage|    |\n+---------------+-----------------+\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/621acf3309b24c538c40824f9af467de\n   :target: https://www.codacy.com/app/AdrienPensart/musicbot?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=AdrienPensart/musicbot&amp;utm_campaign=Badge_Grade\n   :alt: Codacy badge\n.. |coverage| image:: https://github.com/AdrienPensart/musicbot/blob/master/doc/coverage.svg\n   :alt: Coverage badge\n\nDescription\n-----------\nCLI / API / Website to manipulate music and create smart playlists, and play it !\n\nIt uses poetry and pyenv tools to manage project life and docker to test it.\n\nUnder the hood, there is a postgraphile API frontend backed by a postgresql database, it allow us\nto have users and security integrated.\n\nDev Environment\n---------------\n\n.. code-block:: bash\n\n  sudo apt install -y vlc libtag1-dev postgresql-server-dev-all ffmpeg python3-pip docker.io libchromaprint-tools libbz2-dev libsqlite3-dev llvm libncurses5-dev libncursesw5-dev tk-dev liblzma-dev libssl-dev libreadline-dev\n  sudo usermod -aG docker $(whoami)\n\n  git clone https://github.com/AdrienPensart/musicbot.git\n  cd musicbot\n\n  https://pyenv.run | bash\n  pyenv install --verbose $(cat .python-version) -ks\n  pyenv global $(cat .python-version)\n\n  # you should put that at your shell startup\n  eval "$(pyenv init -)"\n\n  python <(curl -sSL https://raw.githubusercontent.com/sdispater/poetry/master/get-poetry.py)\n  poetry install\n\nTesting\n------------\n\n.. code-block:: bash\n\n  poetry run pytest\n  poetry run coverage-badge -f -o doc/coverage.svg\n\nHow to use\n------------\n\n.. code-block:: bash\n\n  poetry run docker-compose build --parallel\n  poetry run docker-compose up\n  musicbot user create --email your_email --password your_password --save\n  musicbot local scan /tests/fixtures/folder1 /tests/fixtures/folder2\n  musicbot local tracks\n\nLinting\n------------\n\n.. code-block:: bash\n\n  poetry run pylint musicbot tests doc\n\nDocumentation\n-------------\n\n.. code-block:: bash\n\n  poetry run rstcheck doc/help.rst\n  poetry run doc/gen.py > README.rst\n  poetry run rstcheck README.rst\n\n\nCommands\n--------\n.. code-block::\n\n  Usage: musicbot [OPTIONS] COMMAND [ARGS]...\n\n    Music swiss knife, new gen.\n\n  Options:\n    Global options: \n      -c, --config FILE                                Config file path  [default: ~/musicbot.ini]\n      -l, --log FILE                                   Log file path\n      -q, --quiet                                      Disable progress bars  [default: False]\n      -t, --timings                                    Set verbosity to info and show execution timings  [default: False]\n    Verbosity: [mutually_exclusive]\n      --debug                                          Same as "--verbosity debug"\n      --info                                           Same as "--verbosity info"\n      --warning                                        Same as "--verbosity warning"\n      --error                                          Same as "--verbosity error"\n      --critical                                       Same as "--verbosity critical"\n      --verbosity [debug|info|warning|error|critical]  Set verbosity level  [default: warning]\n    -V, --version                                      Show the version and exit.\n    -h, --help                                         Show this message and exit.\n\n  Commands:\n    completion    Shell completion\n    folder        Manage folders\n    help          Print help\n    local         Local music management\n    music (file)  Music file\n    spotify       Spotify tool\n    user          User management\n    version       Print version\n    youtube       Youtube tool\n\n\nmusicbot completion\n*******************\n.. code-block::\n\n  Usage: musicbot completion [OPTIONS] COMMAND [ARGS]...\n\n    Shell completion subcommand\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    help                   Print help\n    install                Install the click-completion-command completion\n    show (generate,print)  Show the click-completion-command completion code\n\n\nmusicbot completion install\n***************************\n.. code-block::\n\n  Usage: musicbot completion install [OPTIONS] [[bash|fish|zsh|powershell]] [PATH]\n\n    Auto install shell completion code in your rc file\n\n  Options:\n    -i, --case-insensitive / --no-case-insensitive  Case insensitive completion\n    --append / --overwrite                          Append the completion code to the file\n    -h, --help                                      Show this message and exit.\n\n\nmusicbot completion show\n************************\n.. code-block::\n\n  Usage: musicbot completion show [OPTIONS] [[bash|fish|zsh|powershell]]\n\n    Generate shell code to enable completion\n\n  Options:\n    -i, --case-insensitive / --no-case-insensitive  Case insensitive completion\n    -h, --help                                      Show this message and exit.\n\n\nmusicbot folder\n***************\n.. code-block::\n\n  Usage: musicbot folder [OPTIONS] COMMAND [ARGS]...\n\n    Manage folders\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    find                           Just list music files\n    flac2mp3                       Convert all files in folders to mp3\n    help                           Print help\n    inconsistencies (consistency)  Check music files consistency\n    tracks                         List tracks\n\n\nmusicbot folder find\n********************\n.. code-block::\n\n  Usage: musicbot folder find [OPTIONS] [FOLDERS]...\n\n    Just list music files\n\n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot folder flac2mp3\n************************\n.. code-block::\n\n  Usage: musicbot folder flac2mp3 [OPTIONS] [FOLDERS]...\n\n    Convert all files in folders to mp3\n\n  Options:\n    --folder DIRECTORY     Destination folder\n    --concurrency INTEGER  Number of coroutines  [default: 8]\n    --dry                  Take no real action  [default: False]\n    -h, --help             Show this message and exit.\n\n\nmusicbot folder inconsistencies\n*******************************\n.. code-block::\n\n  Usage: musicbot folder inconsistencies [OPTIONS] [FOLDERS]...\n\n    Check music files consistency\n\n  Options:\n    --dry                                               Take no real action  [default: False]\n    Check options: \n      --checks [no-title|no-artist|no-album|no-genre|no-rating|no-tracknumber|invalid-title|invalid-comment|invalid-path]\n                                                        Consistency tests  [default: no-title, no-artist, no-album, no-genre, no-rating, no-\n                                                        tracknumber, invalid-title, invalid-comment, invalid-path]\n\n      --fix                                             Fix musics\n    -h, --help                                          Show this message and exit.\n\n\nmusicbot folder tracks\n**********************\n.. code-block::\n\n  Usage: musicbot folder tracks [OPTIONS] [FOLDERS]...\n\n    List tracks\n\n  Options:\n    --output [table|json]  Output format  [default: table]\n    -h, --help             Show this message and exit.\n\n\nmusicbot help\n*************\n.. code-block::\n\n  Usage: musicbot help [OPTIONS] [COMMAND]...\n\n    Print help\n\n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local\n**************\n.. code-block::\n\n  Usage: musicbot local [OPTIONS] COMMAND [ARGS]...\n\n    Local music management\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    bests                          Generate bests playlists with some rules\n    clean                          Clean all musics\n    count                          Count musics\n    execute                        Raw query\n    filter                         Print a filter\n    filters                        List filters\n    folders                        List folders\n    help                           Print help\n    inconsistencies (consistency)  Check music consistency\n    load-filters                   Load default filters\n    player (play)                  Music player\n    playlist                       Generate a new playlist\n    rescan                         Clean and load musics\n    scan                           Load musics\n    stats (stat)                   Generate some stats for music collection with filters\n    sync                           Copy selected musics with filters to destination folder\n    watch                          Watch files changes in folders\n\n\nmusicbot local bests\n********************\n.. code-block::\n\n  Usage: musicbot local bests [OPTIONS] FOLDER\n\n    Generate bests playlists with some rules\n\n  Options:\n    --prefix TEXT             Append prefix before each path (implies relative)\n    --suffix TEXT             Append this suffix to playlist name\n    --dry                     Take no real action  [default: False]\n    Auth options: \n      --graphql TEXT          GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT        User email\n      -p, --password TEXT     User password\n      -t, --token TEXT        User token\n    Filter options: \n      --name TEXT             Filter name\n      --limit INTEGER         Fetch a maximum limit of music\n      --youtubes TEXT         Select musics with a youtube link\n      --no-youtubes TEXT      Select musics without youtube link\n      --spotifys TEXT         Select musics with a spotifys link\n      --no-spotifys TEXT      Select musics without spotifys link\n      --formats TEXT          Select musics with file format\n      --no-formats TEXT       Filter musics without format\n      --keywords TEXT         Select musics with keywords\n      --no-keywords TEXT      Filter musics without keywords\n      --artists TEXT          Select musics with artists\n      --no-artists TEXT       Filter musics without artists\n      --albums TEXT           Select musics with albums\n      --no-albums TEXT        Filter musics without albums\n      --titles TEXT           Select musics with titles\n      --no-titles TEXT        Filter musics without titless\n      --genres TEXT           Select musics with genres\n      --no-genres TEXT        Filter musics without genres\n      --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n      --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n      --min-size INTEGER      Minimum file size filter (in bytes)\n      --max-size INTEGER      Maximum file size filter (in bytes)\n      --min-rating FLOAT      Minimum rating  [default: 0.0]\n      --max-rating FLOAT      Maximum rating  [default: 5.0]\n      --relative              Generate relatives paths\n      --shuffle               Randomize selection\n    -h, --help                Show this message and exit.\n\n\nmusicbot local clean\n********************\n.. code-block::\n\n  Usage: musicbot local clean [OPTIONS]\n\n    Clean all musics\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -y, --yes              Confirm action\n    -h, --help             Show this message and exit.\n\n\nmusicbot local count\n********************\n.. code-block::\n\n  Usage: musicbot local count [OPTIONS]\n\n    Count musics\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local execute\n**********************\n.. code-block::\n\n  Usage: musicbot local execute [OPTIONS] QUERY\n\n    Raw query\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local filter\n*********************\n.. code-block::\n\n  Usage: musicbot local filter [OPTIONS] NAME\n\n    Print a filter\n\n  Options:\n    --output [table|json]  Output format  [default: table]\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local filters\n**********************\n.. code-block::\n\n  Usage: musicbot local filters [OPTIONS]\n\n    List filters\n\n  Options:\n    --output [table|json]  Output format  [default: table]\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local folders\n**********************\n.. code-block::\n\n  Usage: musicbot local folders [OPTIONS]\n\n    List folders\n\n  Options:\n    --output [table|json]  Output format  [default: table]\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local inconsistencies\n******************************\n.. code-block::\n\n  Usage: musicbot local inconsistencies [OPTIONS]\n\n    Check music consistency\n\n  Options:\n    Check options: \n      --checks [no-title|no-artist|no-album|no-genre|no-rating|no-tracknumber|invalid-title|invalid-comment|invalid-path]\n                                                        Consistency tests  [default: no-title, no-artist, no-album, no-genre, no-rating, no-\n                                                        tracknumber, invalid-title, invalid-comment, invalid-path]\n\n      --fix                                             Fix musics\n    --dry                                               Take no real action  [default: False]\n    Auth options: \n      --graphql TEXT                                    GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT                                  User email\n      -p, --password TEXT                               User password\n      -t, --token TEXT                                  User token\n    Filter options: \n      --name TEXT                                       Filter name\n      --limit INTEGER                                   Fetch a maximum limit of music\n      --youtubes TEXT                                   Select musics with a youtube link\n      --no-youtubes TEXT                                Select musics without youtube link\n      --spotifys TEXT                                   Select musics with a spotifys link\n      --no-spotifys TEXT                                Select musics without spotifys link\n      --formats TEXT                                    Select musics with file format\n      --no-formats TEXT                                 Filter musics without format\n      --keywords TEXT                                   Select musics with keywords\n      --no-keywords TEXT                                Filter musics without keywords\n      --artists TEXT                                    Select musics with artists\n      --no-artists TEXT                                 Filter musics without artists\n      --albums TEXT                                     Select musics with albums\n      --no-albums TEXT                                  Filter musics without albums\n      --titles TEXT                                     Select musics with titles\n      --no-titles TEXT                                  Filter musics without titless\n      --genres TEXT                                     Select musics with genres\n      --no-genres TEXT                                  Filter musics without genres\n      --min-duration INTEGER                            Minimum duration filter (hours:minutes:seconds)\n      --max-duration INTEGER                            Maximum duration filter (hours:minutes:seconds))\n      --min-size INTEGER                                Minimum file size filter (in bytes)\n      --max-size INTEGER                                Maximum file size filter (in bytes)\n      --min-rating FLOAT                                Minimum rating  [default: 0.0]\n      --max-rating FLOAT                                Maximum rating  [default: 5.0]\n      --relative                                        Generate relatives paths\n      --shuffle                                         Randomize selection\n    -h, --help                                          Show this message and exit.\n\n\nmusicbot local load-filters\n***************************\n.. code-block::\n\n  Usage: musicbot local load-filters [OPTIONS]\n\n    Load default filters\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local player\n*********************\n.. code-block::\n\n  Usage: musicbot local player [OPTIONS]\n\n    Music player\n\n  Options:\n    Auth options: \n      --graphql TEXT          GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT        User email\n      -p, --password TEXT     User password\n      -t, --token TEXT        User token\n    Filter options: \n      --name TEXT             Filter name\n      --limit INTEGER         Fetch a maximum limit of music\n      --youtubes TEXT         Select musics with a youtube link\n      --no-youtubes TEXT      Select musics without youtube link\n      --spotifys TEXT         Select musics with a spotifys link\n      --no-spotifys TEXT      Select musics without spotifys link\n      --formats TEXT          Select musics with file format\n      --no-formats TEXT       Filter musics without format\n      --keywords TEXT         Select musics with keywords\n      --no-keywords TEXT      Filter musics without keywords\n      --artists TEXT          Select musics with artists\n      --no-artists TEXT       Filter musics without artists\n      --albums TEXT           Select musics with albums\n      --no-albums TEXT        Filter musics without albums\n      --titles TEXT           Select musics with titles\n      --no-titles TEXT        Filter musics without titless\n      --genres TEXT           Select musics with genres\n      --no-genres TEXT        Filter musics without genres\n      --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n      --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n      --min-size INTEGER      Minimum file size filter (in bytes)\n      --max-size INTEGER      Maximum file size filter (in bytes)\n      --min-rating FLOAT      Minimum rating  [default: 0.0]\n      --max-rating FLOAT      Maximum rating  [default: 5.0]\n      --relative              Generate relatives paths\n      --shuffle               Randomize selection\n    -h, --help                Show this message and exit.\n\n\nmusicbot local playlist\n***********************\n.. code-block::\n\n  Usage: musicbot local playlist [OPTIONS] [PATH]\n\n    Generate a new playlist\n\n  Options:\n    --dry                      Take no real action  [default: False]\n    --output [json|m3u|table]  Output format  [default: table]\n    Auth options: \n      --graphql TEXT           GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT         User email\n      -p, --password TEXT      User password\n      -t, --token TEXT         User token\n    Filter options: \n      --name TEXT              Filter name\n      --limit INTEGER          Fetch a maximum limit of music\n      --youtubes TEXT          Select musics with a youtube link\n      --no-youtubes TEXT       Select musics without youtube link\n      --spotifys TEXT          Select musics with a spotifys link\n      --no-spotifys TEXT       Select musics without spotifys link\n      --formats TEXT           Select musics with file format\n      --no-formats TEXT        Filter musics without format\n      --keywords TEXT          Select musics with keywords\n      --no-keywords TEXT       Filter musics without keywords\n      --artists TEXT           Select musics with artists\n      --no-artists TEXT        Filter musics without artists\n      --albums TEXT            Select musics with albums\n      --no-albums TEXT         Filter musics without albums\n      --titles TEXT            Select musics with titles\n      --no-titles TEXT         Filter musics without titless\n      --genres TEXT            Select musics with genres\n      --no-genres TEXT         Filter musics without genres\n      --min-duration INTEGER   Minimum duration filter (hours:minutes:seconds)\n      --max-duration INTEGER   Maximum duration filter (hours:minutes:seconds))\n      --min-size INTEGER       Minimum file size filter (in bytes)\n      --max-size INTEGER       Maximum file size filter (in bytes)\n      --min-rating FLOAT       Minimum rating  [default: 0.0]\n      --max-rating FLOAT       Maximum rating  [default: 5.0]\n      --relative               Generate relatives paths\n      --shuffle                Randomize selection\n    -h, --help                 Show this message and exit.\n\n\nmusicbot local rescan\n*********************\n.. code-block::\n\n  Usage: musicbot local rescan [OPTIONS] [FOLDERS]...\n\n    Clean and load musics\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local scan\n*******************\n.. code-block::\n\n  Usage: musicbot local scan [OPTIONS] [FOLDERS]...\n\n    Load musics\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot local stats\n********************\n.. code-block::\n\n  Usage: musicbot local stats [OPTIONS]\n\n    Generate some stats for music collection with filters\n\n  Options:\n    --output [table|json]     Output format  [default: table]\n    Auth options: \n      --graphql TEXT          GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT        User email\n      -p, --password TEXT     User password\n      -t, --token TEXT        User token\n    Filter options: \n      --name TEXT             Filter name\n      --limit INTEGER         Fetch a maximum limit of music\n      --youtubes TEXT         Select musics with a youtube link\n      --no-youtubes TEXT      Select musics without youtube link\n      --spotifys TEXT         Select musics with a spotifys link\n      --no-spotifys TEXT      Select musics without spotifys link\n      --formats TEXT          Select musics with file format\n      --no-formats TEXT       Filter musics without format\n      --keywords TEXT         Select musics with keywords\n      --no-keywords TEXT      Filter musics without keywords\n      --artists TEXT          Select musics with artists\n      --no-artists TEXT       Filter musics without artists\n      --albums TEXT           Select musics with albums\n      --no-albums TEXT        Filter musics without albums\n      --titles TEXT           Select musics with titles\n      --no-titles TEXT        Filter musics without titless\n      --genres TEXT           Select musics with genres\n      --no-genres TEXT        Filter musics without genres\n      --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n      --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n      --min-size INTEGER      Minimum file size filter (in bytes)\n      --max-size INTEGER      Maximum file size filter (in bytes)\n      --min-rating FLOAT      Minimum rating  [default: 0.0]\n      --max-rating FLOAT      Maximum rating  [default: 5.0]\n      --relative              Generate relatives paths\n      --shuffle               Randomize selection\n    -h, --help                Show this message and exit.\n\n\nmusicbot local sync\n*******************\n.. code-block::\n\n  Usage: musicbot local sync [OPTIONS] DESTINATION\n\n    Copy selected musics with filters to destination folder\n\n  Options:\n    --dry                     Take no real action  [default: False]\n    Auth options: \n      --graphql TEXT          GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT        User email\n      -p, --password TEXT     User password\n      -t, --token TEXT        User token\n    Filter options: \n      --name TEXT             Filter name\n      --limit INTEGER         Fetch a maximum limit of music\n      --youtubes TEXT         Select musics with a youtube link\n      --no-youtubes TEXT      Select musics without youtube link\n      --spotifys TEXT         Select musics with a spotifys link\n      --no-spotifys TEXT      Select musics without spotifys link\n      --formats TEXT          Select musics with file format\n      --no-formats TEXT       Filter musics without format\n      --keywords TEXT         Select musics with keywords\n      --no-keywords TEXT      Filter musics without keywords\n      --artists TEXT          Select musics with artists\n      --no-artists TEXT       Filter musics without artists\n      --albums TEXT           Select musics with albums\n      --no-albums TEXT        Filter musics without albums\n      --titles TEXT           Select musics with titles\n      --no-titles TEXT        Filter musics without titless\n      --genres TEXT           Select musics with genres\n      --no-genres TEXT        Filter musics without genres\n      --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n      --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n      --min-size INTEGER      Minimum file size filter (in bytes)\n      --max-size INTEGER      Maximum file size filter (in bytes)\n      --min-rating FLOAT      Minimum rating  [default: 0.0]\n      --max-rating FLOAT      Maximum rating  [default: 5.0]\n      --relative              Generate relatives paths\n      --shuffle               Randomize selection\n    -h, --help                Show this message and exit.\n\n\nmusicbot local watch\n********************\n.. code-block::\n\n  Usage: musicbot local watch [OPTIONS]\n\n    Watch files changes in folders\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot music\n**************\n.. code-block::\n\n  Usage: musicbot music [OPTIONS] COMMAND [ARGS]...\n\n    Music file\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    add-keywords                   Add keywords to music\n    delete-keywords                Delete keywords to music\n    fingerprint                    Print music fingerprint\n    flac2mp3                       Convert flac music to mp3\n    help                           Print help\n    inconsistencies (consistency)  Check music consistency\n    set-tags                       Set music title\n    tags                           Print music tags\n\n\nmusicbot music add-keywords\n***************************\n.. code-block::\n\n  Usage: musicbot music add-keywords [OPTIONS] PATH [KEYWORDS]...\n\n    Add keywords to music\n\n  Options:\n    --dry       Take no real action  [default: False]\n    -h, --help  Show this message and exit.\n\n\nmusicbot music delete-keywords\n******************************\n.. code-block::\n\n  Usage: musicbot music delete-keywords [OPTIONS] PATH [KEYWORDS]...\n\n    Delete keywords to music\n\n  Options:\n    --dry       Take no real action  [default: False]\n    -h, --help  Show this message and exit.\n\n\nmusicbot music fingerprint\n**************************\n.. code-block::\n\n  Usage: musicbot music fingerprint [OPTIONS] PATH\n\n    Print music fingerprint\n\n  Options:\n    --acoustid-api-key TEXT  AcoustID API Key\n    -h, --help               Show this message and exit.\n\n\nmusicbot music flac2mp3\n***********************\n.. code-block::\n\n  Usage: musicbot music flac2mp3 [OPTIONS] PATH\n\n    Convert flac music to mp3\n\n  Options:\n    --folder DIRECTORY  Destination folder\n    --dry               Take no real action  [default: False]\n    -h, --help          Show this message and exit.\n\n\nmusicbot music inconsistencies\n******************************\n.. code-block::\n\n  Usage: musicbot music inconsistencies [OPTIONS] PATH\n\n    Check music consistency\n\n  Options:\n    --folder DIRECTORY                                  Destination folder\n    --dry                                               Take no real action  [default: False]\n    Check options: \n      --checks [no-title|no-artist|no-album|no-genre|no-rating|no-tracknumber|invalid-title|invalid-comment|invalid-path]\n                                                        Consistency tests  [default: no-title, no-artist, no-album, no-genre, no-rating, no-\n                                                        tracknumber, invalid-title, invalid-comment, invalid-path]\n\n      --fix                                             Fix musics\n    -h, --help                                          Show this message and exit.\n\n\nmusicbot music set-tags\n***********************\n.. code-block::\n\n  Usage: musicbot music set-tags [OPTIONS] PATH\n\n    Set music title\n\n  Options:\n    --dry              Take no real action  [default: False]\n    Music options: \n      --keywords TEXT  Keywords\n      --artist TEXT    Artist\n      --album TEXT     Album\n      --title TEXT     Title\n      --genre TEXT     Genre\n      --number TEXT    Track number\n      --rating TEXT    Rating\n    -h, --help         Show this message and exit.\n\n\nmusicbot music tags\n*******************\n.. code-block::\n\n  Usage: musicbot music tags [OPTIONS] PATH\n\n    Print music tags\n\n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot spotify\n****************\n.. code-block::\n\n  Usage: musicbot spotify [OPTIONS] COMMAND [ARGS]...\n\n    Spotify tool\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    diff       Diff between local and spotify\n    help       Print help\n    playlist   Show playlist\n    playlists  List playlists\n    tracks     Show tracks\n\n\nmusicbot spotify diff\n*********************\n.. code-block::\n\n  Usage: musicbot spotify diff [OPTIONS]\n\n    Diff between local and spotify\n\n  Options:\n    Auth options: \n      --graphql TEXT             GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT           User email\n      -p, --password TEXT        User password\n      -t, --token TEXT           User token\n    Spotify options: \n      --username TEXT            Spotify username\n      --client-id TEXT           Spotify client ID\n      --client-secret TEXT       Spotify client secret\n      --token TEXT               Spotify token\n      --cache-path FILE          Spotify cache path\n      --scopes TEXT              Spotify scopes\n      --redirect-uri TEXT        Spotify redirect URI\n    --output [table|json]        Output format  [default: table]\n    --min-threshold FLOAT RANGE  Minimum distance threshold\n    --max-threshold FLOAT RANGE  Maximum distance threshold\n    -h, --help                   Show this message and exit.\n\n\nmusicbot spotify playlist\n*************************\n.. code-block::\n\n  Usage: musicbot spotify playlist [OPTIONS] NAME\n\n    Show playlist\n\n  Options:\n    Spotify options: \n      --username TEXT       Spotify username\n      --client-id TEXT      Spotify client ID\n      --client-secret TEXT  Spotify client secret\n      --token TEXT          Spotify token\n      --cache-path FILE     Spotify cache path\n      --scopes TEXT         Spotify scopes\n      --redirect-uri TEXT   Spotify redirect URI\n    --output [table|json]   Output format  [default: table]\n    -h, --help              Show this message and exit.\n\n\nmusicbot spotify playlists\n**************************\n.. code-block::\n\n  Usage: musicbot spotify playlists [OPTIONS]\n\n    List playlists\n\n  Options:\n    Spotify options: \n      --username TEXT       Spotify username\n      --client-id TEXT      Spotify client ID\n      --client-secret TEXT  Spotify client secret\n      --token TEXT          Spotify token\n      --cache-path FILE     Spotify cache path\n      --scopes TEXT         Spotify scopes\n      --redirect-uri TEXT   Spotify redirect URI\n    -h, --help              Show this message and exit.\n\n\nmusicbot spotify tracks\n***********************\n.. code-block::\n\n  Usage: musicbot spotify tracks [OPTIONS]\n\n    Show tracks\n\n  Options:\n    Spotify options: \n      --username TEXT       Spotify username\n      --client-id TEXT      Spotify client ID\n      --client-secret TEXT  Spotify client secret\n      --token TEXT          Spotify token\n      --cache-path FILE     Spotify cache path\n      --scopes TEXT         Spotify scopes\n      --redirect-uri TEXT   Spotify redirect URI\n    --output [table|json]   Output format  [default: table]\n    -h, --help              Show this message and exit.\n\n\nmusicbot user\n*************\n.. code-block::\n\n  Usage: musicbot user [OPTIONS] COMMAND [ARGS]...\n\n    User management\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    help                        Print help\n    list                        List users (admin)\n    login (token)               Authenticate user\n    register (add,create,new)   Register a new user\n    unregister (delete,remove)  Remove a user\n\n\nmusicbot user list\n******************\n.. code-block::\n\n  Usage: musicbot user list [OPTIONS]\n\n    List users (admin)\n\n  Options:\n    --output [table|json]            Output format  [default: table]\n    Admin options: \n      --graphql-admin TEXT           GraphQL endpoint  [default: http://127.0.0.1:5001/graphql]\n      --graphql-admin-user TEXT      GraphQL admin user (basic auth)\n      --graphql-admin-password TEXT  GraphQL admin password (basic auth)\n    -h, --help                       Show this message and exit.\n\n\nmusicbot user login\n*******************\n.. code-block::\n\n  Usage: musicbot user login [OPTIONS]\n\n    Authenticate user\n\n  Options:\n    -s, --save             Save to config file  [default: False]\n    User options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n    -h, --help             Show this message and exit.\n\n\nmusicbot user register\n**********************\n.. code-block::\n\n  Usage: musicbot user register [OPTIONS]\n\n    Register a new user\n\n  Options:\n    -s, --save             Save to config file  [default: False]\n    Register options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      --first-name TEXT    User first name\n      --last-name TEXT     User last name\n    -h, --help             Show this message and exit.\n\n\nmusicbot user unregister\n************************\n.. code-block::\n\n  Usage: musicbot user unregister [OPTIONS]\n\n    Remove a user\n\n  Options:\n    Auth options: \n      --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n      -e, --email TEXT     User email\n      -p, --password TEXT  User password\n      -t, --token TEXT     User token\n    -h, --help             Show this message and exit.\n\n\nmusicbot version\n****************\n.. code-block::\n\n  Usage: musicbot version [OPTIONS]\n\n    Print version\n\n    Equivalent : -V\n\n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot youtube\n****************\n.. code-block::\n\n  Usage: musicbot youtube [OPTIONS] COMMAND [ARGS]...\n\n    Youtube tool\n\n  Options:\n    -h, --help  Show this message and exit.\n\n  Commands:\n    download     Download a youtube link with artist and title\n    find         Search a youtube link with artist and title\n    fingerprint  Fingerprint a youtube video\n    help         Print help\n    search       Search a youtube link with artist and title\n\n\nmusicbot youtube download\n*************************\n.. code-block::\n\n  Usage: musicbot youtube download [OPTIONS] ARTIST TITLE\n\n    Download a youtube link with artist and title\n\n  Options:\n    --path TEXT\n    -h, --help   Show this message and exit.\n\n\nmusicbot youtube find\n*********************\n.. code-block::\n\n  Usage: musicbot youtube find [OPTIONS] PATH\n\n    Search a youtube link with artist and title\n\n  Options:\n    --acoustid-api-key TEXT  AcoustID API Key\n    -h, --help               Show this message and exit.\n\n\nmusicbot youtube fingerprint\n****************************\n.. code-block::\n\n  Usage: musicbot youtube fingerprint [OPTIONS] URL\n\n    Fingerprint a youtube video\n\n  Options:\n    --acoustid-api-key TEXT  AcoustID API Key\n    -h, --help               Show this message and exit.\n\n\nmusicbot youtube search\n***********************\n.. code-block::\n\n  Usage: musicbot youtube search [OPTIONS] ARTIST TITLE\n\n    Search a youtube link with artist and title\n\n  Options:\n    -h, --help  Show this message and exit.\n',
    'author': 'Adrien Pensart',
    'author_email': 'crunchengine@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



import pathlib
import logging
from os import scandir, walk, path, DirEntry
from typing import Iterator, Iterable, Tuple
import humanfriendly  # type: ignore

logger = logging.getLogger(__name__)

output_types = ["list", "json"]
except_directories = ['.Spotlight-V100', '.zfs', 'Android', 'LOST.DIR']
default_output_type = 'json'


def ensure(path: str) -> str:
    p = pathlib.Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    return str(p)


def bytes_to_human(b: int) -> str:
    return str(humanfriendly.format_size(b))


def empty_dirs(root_dir: str, recursive: bool = True) -> Iterator[str]:
    dirs_list = []
    for root, dirs, files in walk(root_dir, topdown=False):
        if recursive:
            all_subs_empty = True
            for sub in dirs:
                full_sub = path.join(root, sub)
                if full_sub not in dirs_list:
                    all_subs_empty = False
                    break
        else:
            all_subs_empty = (not dirs)
        if all_subs_empty and not files:
            dirs_list.append(root)
            yield root


def find_files(directories: Iterable[str], supported_formats: Iterable[str]) -> Iterator[Tuple[str, str]]:
    directories = [path.abspath(d) for d in directories]
    for directory in directories:
        for root, _, files in walk(directory):
            if any(e in root for e in except_directories):
                continue
            for basename in files:
                filename = path.join(root, basename)
                if filename.endswith(tuple(supported_formats)):
                    yield (directory, filename)


def scantree(path: str, supported_formats: Iterable[str]) -> Iterator[DirEntry]:
    try:
        if '/.' in path:
            return
        with scandir(path) as it:
            for entry in it:
                if entry.is_file():
                    if entry.name.endswith(tuple(supported_formats)):
                        yield entry
                elif entry.is_dir(follow_symlinks=False):
                    yield from scantree(entry.path, supported_formats)
    except PermissionError as e:
        logger.error(e)


def filecount(path: str, supported_formats: Iterable[str]) -> int:
    return len(list(scantree(path, supported_formats)))


def all_files(directory: str) -> Iterable[str]:
    for root, _, files in walk(directory):
        if any(e in root for e in except_directories):
            logger.debug(f"Invalid path {root}")
            continue
        for basename in files:
            yield path.join(root, basename)


default_checks = [
    'keywords',
    'strict_title',
    'title',
    'path',
    'genre',
    'album',
    'artist',
    'rating',
    'number'
]



'''Musicbot CLI'''
import logging
import click
from click_skeleton import skeleton, sensible_context_settings, add_options
from musicbot.global_options import options

PROG_NAME = "musicbot"
__version__ = "0.7.4"
CONTEXT_SETTINGS = sensible_context_settings(PROG_NAME, __version__, auto_envvar_prefix='MB')
logger = logging.getLogger(PROG_NAME)


@skeleton(context_settings=CONTEXT_SETTINGS)
@click.pass_context
@add_options(options)
def main_cli(ctx, **kwargs):
    """Music swiss knife, new gen."""
    from musicbot import commands
    from musicbot.config import config as config_obj
    config_obj.set(**kwargs)
    ctx.obj.config = config_obj
    logger.debug(f"Imported commands: {commands.modules}")
